@@ -106,7 +106,7 @@ typedef struct {
  *      -# receives two references to elements a and b
  *      -# returns {<0, 0, >0} if (a > b), (a == b), (a < b) respectively
  *
- * It is responsability of the function to handle possible NULL values.
+ * It is responsibility of the function to handle possible NULL values.
  */
 typedef int (*element_comparator)(const void *a, const void *b);
 
@@ -118,19 +118,19 @@ typedef int (*element_comparator)(const void *a, const void *b);
  *      -# receives a reference to some indicator data
  *      -# returns non-0 if the element matches the indicator, 0 otherwise
  *
- * It is responsability of the function to handle possible NULL values in any
+ * It is responsibility of the function to handle possible NULL values in any
  * argument.
  */
 typedef int (*element_seeker)(const void *el, const void *indicator);
 
 /**
- * an element lenght meter.
+ * an element length meter.
  *
  * An element meter is a function that:
  *      -# receives the reference to an element el
  *      -# returns its size in bytes
  *
- * It is responsability of the function to handle possible NULL values.
+ * It is responsibility of the function to handle possible NULL values.
  */
 typedef size_t (*element_meter)(const void *el);
 
@@ -141,7 +141,7 @@ typedef size_t (*element_meter)(const void *el);
  *      -# receives the reference to an element el
  *      -# returns a hash value for el
  *
- * It is responsability of the function to handle possible NULL values.
+ * It is responsibility of the function to handle possible NULL values.
  */
 typedef list_hash_t (*element_hash_computer)(const void *el);
 
@@ -151,7 +151,7 @@ typedef list_hash_t (*element_hash_computer)(const void *el);
  * A serializer function is one that gets a reference to an element,
  * and returns a reference to a buffer that contains its serialization
  * along with the length of this buffer.
- * It is responsability of the function to handle possible NULL values,
+ * It is responsibility of the function to handle possible NULL values,
  * returning a NULL buffer and a 0 buffer length.
  *
  * These functions have 3 goals:
@@ -163,7 +163,7 @@ typedef list_hash_t (*element_hash_computer)(const void *el);
  * @param serialize_buffer  reference to fill with the length of the buffer
  * @return                  reference to the buffer with the serialized data
  */
-typedef void *(*element_serializer)(const void *simclist_restrict el, uint32_t *simclist_restrict serializ_len);
+typedef void *(*element_serializer)(const void *simclist_restrict el, uint32_t *simclist_restrict serialize_buffer);
 
 /**
  * a function for un-serializing an element.
@@ -290,7 +290,7 @@ int list_attributes_seeker(list_t *simclist_restrict l, element_seeker seeker_fu
  * the list by its actual data is not free()d. With this option, every
  * deletion causes element data to be freed.
  *
- * It is responsability of this function to correctly handle NULL values, if
+ * It is responsibility of this function to correctly handle NULL values, if
  * NULL elements are inserted into the list.
  *
  * @param l             list to operate
@@ -470,7 +470,7 @@ int list_insert_at(list_t *simclist_restrict l, const void *data, unsigned int p
  * expunge the first found given element from the list.
  *
  * Inspects the given list looking for the given element; if the element
- * is found, it is removed. Only the first occurence is removed.
+ * is found, it is removed. Only the first occurrence is removed.
  * If a comparator function was not set, elements are compared by reference.
  * Otherwise, the comparator is used to match the element.
  *
@@ -505,7 +505,7 @@ int list_delete_range(list_t *simclist_restrict l, unsigned int posstart, unsign
 /**
  * clear all the elements off of the list.
  *
- * The element datums will not be freed.
+ * The element data will not be freed.
  *
  * @see list_delete_range()
  * @see list_size()
@@ -595,7 +595,7 @@ int list_contains(const list_t *simclist_restrict l, const void *data);
  * user-provided list object, which must be different from both the
  * lists to concatenate. Attributes from the original lists are not
  * cloned.
- * The destination list referred is threated as virgin room: if it
+ * The destination list referred is treated as virgin room: if it
  * is an existing list containing elements, memory leaks will happen.
  * It is OK to specify the same list twice as source, for "doubling"
  * it in the destination.
@@ -672,7 +672,7 @@ int list_hash(const list_t *simclist_restrict l, list_hash_t *simclist_restrict
 
 #ifdef SIMCLIST_DUMPRESTORE
 /**
- * get meta informations on a list dump on filedescriptor.
+ * get meta information on a list dump on filedescriptor.
  *
  * [ advanced function ]
  *
@@ -689,7 +689,7 @@ int list_hash(const list_t *simclist_restrict l, list_hash_t *simclist_restrict
 int list_dump_getinfo_filedescriptor(int fd, list_dump_info_t *simclist_restrict info);
 
 /**
- * get meta informations on a list dump on file.
+ * get meta information on a list dump on file.
  *
  * [ advanced function ]
  *
