@@ -255,11 +255,13 @@ static int gids_get_DO(sc_card_t* card, int fileIdentifier, int dataObjectIdenti
 	if (!p) {
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_FILE_NOT_FOUND);
 	}
-	if (datasize > *responselen) {
-		LOG_FUNC_RETURN(card->ctx, SC_ERROR_BUFFER_TOO_SMALL);
+	if (response && responselen) {
+		if (datasize > *responselen) {
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_BUFFER_TOO_SMALL);
+		}
+		memcpy(response, p, datasize);
+		*responselen = datasize;
 	}
-	memcpy(response, p, datasize);
-	*responselen = datasize;
 	return SC_SUCCESS;
 }
 
@@ -1029,7 +1031,6 @@ static int gids_get_pin_policy(struct sc_card *card, struct sc_pin_cmd_data *dat
 	LOG_TEST_RET(card->ctx, r, "gids_get_pin_status failed");
 	data->pin1.max_length = 16;
 	data->pin1.min_length = 4;
-	data->pin1.stored_length = 0;
 	data->pin1.encoding = SC_PIN_ENCODING_ASCII;
 	data->pin1.offset = 5;
 	data->pin1.logged_in = SC_PIN_STATE_UNKNOWN;
