@@ -1,7 +1,7 @@
 /*
  * card-myeid.c
  *
- * Copyright (C) 2008-2009 Aventra Ltd.
+ * Copyright (C) 2008-2019 Aventra Ltd.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -30,26 +30,37 @@
 #include "cardctl.h"
 #include "types.h"
 
-#define LOAD_KEY_MODULUS		0x80
-#define LOAD_KEY_PUBLIC_EXPONENT	0x81
-#define LOAD_KEY_PRIME_P		0x83
-#define LOAD_KEY_PRIME_Q		0x84
-#define LOAD_KEY_DP1			0x85
-#define LOAD_KEY_DQ1			0x86
-#define LOAD_KEY_INVQ			0x87
-#define LOAD_KEY_MODE_EC_PRIV		0x87
-#define LOAD_KEY_MODE_EC_PUB		0x86
+/* Low byte is the MyEID card's key type specific component ID. High byte is used
+ * internally for key type, so myeid_loadkey() is aware of the exact component. */
+#define LOAD_KEY_MODULUS		0x0080
+#define LOAD_KEY_PUBLIC_EXPONENT	0x0081
+#define LOAD_KEY_PRIME_P		0x0083
+#define LOAD_KEY_PRIME_Q		0x0084
+#define LOAD_KEY_DP1			0x0085
+#define LOAD_KEY_DQ1			0x0086
+#define LOAD_KEY_INVQ			0x0087
+#define LOAD_KEY_EC_PUBLIC		0x1086
+#define LOAD_KEY_EC_PRIVATE		0x1087
+#define LOAD_KEY_SYMMETRIC		0x20a0
 
-#define LOAD_KEY_EC_PRIVATE	0x97
-#define LOAD_KEY_EC_PUBLIC	0x96
-
-#define MYEID_STATE_CREATION	0x01
-#define MYEID_STATE_ACTIVATED	0x07
-
-#define MYEID_INFINEON_CHIP_ATR		0x04
 #define MYEID_CARD_NAME_MAX_LEN		100
 
+/* The following flags define the features supported by the card currently in use.
+  They are used in 'card_supported_features' field in myeid_card_caps struct */
+#define MYEID_CARD_CAP_RSA		0x01
+#define MYEID_CARD_CAP_3DES		0x02
+#define MYEID_CARD_CAP_AES		0x04
+#define MYEID_CARD_CAP_ECC		0x08
+#define MYEID_CARD_CAP_GRIDPIN		0x10
+#define MYEID_CARD_CAP_PIV_EMU		0x20
+
+#define MYEID_MAX_APDU_DATA_LEN		0xFF
+#define MYEID_MAX_RSA_KEY_LEN		4096
+
+#define MYEID_MAX_EXT_APDU_BUFFER_SIZE	(MYEID_MAX_RSA_KEY_LEN/8+16)
+
 static const char *myeid_card_name = "MyEID";
+static const char *oseid_card_name = "OsEID";
 static char card_name_buf[MYEID_CARD_NAME_MAX_LEN];
 
 static struct sc_card_operations myeid_ops;
@@ -62,27 +73,28 @@ static struct sc_card_driver myeid_drv = {
 	NULL
 };
 
-static const char *myeid_atrs[] = {
-	"3B:F5:18:00:FF:81:31:FE:45:4D:79:45:49:44:65",
-	"3B:F5:18:00:00:81:31:FE:45:4D:79:45:49:44:9A",
-	"3B:85:80:01:4D:79:45:49:44:78",
-	"3B:89:80:01:09:38:33:B1:4D:79:45:49:44:4C",
-	"3B:F5:96:00:00:80:31:FE:45:4D:79:45:49:44:15", /* Infineon's chip */
-	"3B:F5:96:00:00:81:31:FE:45:4D:79:45:49:44:14",
-	NULL
-};
-
 typedef struct myeid_private_data {
 	int card_state;
 
 	unsigned short change_counter;
+	unsigned char cap_chaining;
 	/* the driver sets sec_env pointer in myeid_set_security_env and
 	 it is used immediately in myeid_decipher to differentiate between RSA decryption and
 	 ECDH key agreement. Note that this pointer is usually not valid
 	 after this pair of calls and must not be used elsewhere. */
 	const struct sc_security_env* sec_env;
+	int disable_hw_pkcs1_padding;
 } myeid_private_data_t;
 
+typedef struct myeid_card_caps {
+	unsigned char card_caps_ver;
+	unsigned short card_supported_features;
+	unsigned short max_rsa_key_length;
+	unsigned short max_des_key_length;
+	unsigned short max_aes_key_length;
+	unsigned short max_ecc_key_length;
+} myeid_card_caps_t;
+
 static struct myeid_supported_ec_curves {
 	char *curve_name;
 	struct sc_object_id curve_oid;
@@ -97,30 +109,87 @@ static struct myeid_supported_ec_curves {
 };
 
 static int myeid_get_info(struct sc_card *card, u8 *rbuf, size_t buflen);
+static int myeid_get_card_caps(struct sc_card *card, myeid_card_caps_t* card_caps);
 
 static int myeid_match_card(struct sc_card *card)
 {
-	int i, match = -1;
-
-	for (i = 0; myeid_atrs[i] != NULL; i++)
-	{
-		u8 defatr[SC_MAX_ATR_SIZE];
-		size_t len = sizeof(defatr);
-		const char *atrp = myeid_atrs[i];
-
-		if (sc_hex_to_bin(atrp, defatr, &len))
-			continue;
-		if (len != card->atr.len)
-			continue;
-		if (memcmp(card->atr.value, defatr, len) != 0)
-			continue;
-		match = i;
-		break;
+	size_t len = card->reader->atr_info.hist_bytes_len;
+	/* Normally the historical bytes are exactly "MyEID", but there might
+	 * be some historic units which have a small prefix byte sequence. */
+	if (len >= 5) {
+		if (!memcmp(&card->reader->atr_info.hist_bytes[len - 5], "MyEID", 5)) {
+			sc_log(card->ctx, "Matched MyEID card");
+			card->type = SC_CARD_TYPE_MYEID_GENERIC;
+			return 1;
+		}
+		/* The software implementation of MyEID is identified by OsEID bytes */
+		if (!memcmp(&card->reader->atr_info.hist_bytes[len - 5], "OsEID", 5)) {
+			sc_log(card->ctx, "Matched OsEID card");
+			card->type = SC_CARD_TYPE_MYEID_OSEID;
+			return 1;
+		}
 	}
-	if (match == -1)
-		return 0;
+	return 0;
+}
 
-	return 1;
+static int
+myeid_select_aid(struct sc_card *card, struct sc_aid *aid, unsigned char *out, size_t *out_len)
+{
+	struct sc_apdu apdu;
+	unsigned char apdu_resp[SC_MAX_APDU_BUFFER_SIZE];
+	int rv;
+
+	/* Select application (deselect previously selected application) */
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0xA4, 0x04, 0x00);
+	apdu.lc = aid->len;
+	apdu.data = aid->value;
+	apdu.datalen = aid->len;
+	apdu.resplen = sizeof(apdu_resp);
+	apdu.resp = apdu_resp;
+
+	rv = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_RET(card->ctx, rv, "APDU transmit failed");
+	rv = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, rv, "Cannot select AID");
+
+	if (*out_len > 0) {
+		if (*out_len < apdu.resplen)
+			LOG_TEST_RET(card->ctx, SC_ERROR_BUFFER_TOO_SMALL, "Cannot select AID - response buffer too small.");
+		if (out == NULL)
+			LOG_TEST_RET(card->ctx, SC_ERROR_INVALID_ARGUMENTS, "Cannot select AID - invalid arguments.");
+		memcpy(out, apdu.resp, apdu.resplen);
+		*out_len = apdu.resplen;
+	}
+
+	return SC_SUCCESS;
+}
+
+static int myeid_load_options(sc_context_t *ctx, myeid_private_data_t *priv)
+{
+	int r;
+	size_t i, j;
+	scconf_block **found_blocks, *block;
+
+	if (!ctx || !priv) {
+		r = SC_ERROR_INTERNAL;
+		goto err;
+	}
+	priv->disable_hw_pkcs1_padding = 0;
+	for (i = 0; ctx->conf_blocks[i]; i++) {
+		found_blocks = scconf_find_blocks(ctx->conf, ctx->conf_blocks[i],
+				"card_driver", "myeid");
+		if (!found_blocks)
+			continue;
+		for (j = 0, block = found_blocks[j]; block; j++, block = found_blocks[j]) {
+			priv->disable_hw_pkcs1_padding = scconf_get_int(block, "disable_hw_pkcs1_padding", 0);
+			sc_log(ctx,"Found config option: disable_hw_pkcs1_padding = %d\n", priv->disable_hw_pkcs1_padding);
+		}
+		free(found_blocks);
+	}
+	r = SC_SUCCESS;
+
+err:
+	return r;
 }
 
 static int myeid_init(struct sc_card *card)
@@ -129,36 +198,65 @@ static int myeid_init(struct sc_card *card)
 	myeid_private_data_t *priv;
 	u8 appletInfo[20];
 	size_t appletInfoLen;
-	int r;
-	int largeEccKeys = 0;
-	u8 defatr[SC_MAX_ATR_SIZE];
-	size_t len = sizeof(defatr);
-	const char *atrp = myeid_atrs[MYEID_INFINEON_CHIP_ATR];
+	myeid_card_caps_t card_caps;
+	size_t resp_len = 0;
+	static struct sc_aid myeid_aid = { "\xA0\x00\x00\x00\x63\x50\x4B\x43\x53\x2D\x31\x35", 0x0C };
+	int rv = 0;
+	void *old_drv_data = card->drv_data;
 
 	LOG_FUNC_CALLED(card->ctx);
 
-	card->name = myeid_card_name;
+	switch (card->type) {
+	case SC_CARD_TYPE_MYEID_OSEID:
+		card->name = oseid_card_name;
+		break;
+	case SC_CARD_TYPE_MYEID_GENERIC:
+		card->name = myeid_card_name;
+		break;
+	default:
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_CARD);
+	}
 
 	priv = calloc(1, sizeof(myeid_private_data_t));
 
 	if (!priv)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
 
+	rv = myeid_load_options (card->ctx, priv);
+	LOG_TEST_GOTO_ERR(card->ctx, rv, "Unable to read options from opensc.conf");
+
 	priv->card_state = SC_FILE_STATUS_CREATION;
 	card->drv_data = priv;
 
+	/* Ensure that the MyEID applet is selected. */	
+	rv = myeid_select_aid(card, &myeid_aid, NULL, &resp_len);
+	LOG_TEST_GOTO_ERR(card->ctx, rv, "Failed to select MyEID applet.");
+
 	/* find out MyEID version */
 
 	appletInfoLen = 20;
 
-	r = myeid_get_info(card, appletInfo, appletInfoLen);
-
-	LOG_TEST_RET(card->ctx, r, "Failed to get MyEID applet information.");
+	if (0 > myeid_get_info(card, appletInfo, appletInfoLen))
+		LOG_TEST_GOTO_ERR(card->ctx, SC_ERROR_INVALID_CARD, "Failed to get MyEID applet information.");
 
 	priv->change_counter = appletInfo[19] | appletInfo[18] << 8;
 
-	flags = SC_ALGORITHM_RSA_RAW | SC_ALGORITHM_RSA_PAD_PKCS1 | SC_ALGORITHM_ONBOARD_KEY_GEN;
-	flags |= SC_ALGORITHM_RSA_HASH_NONE | SC_ALGORITHM_RSA_HASH_SHA1;
+	memset(&card_caps, 0, sizeof(myeid_card_caps_t));
+	card_caps.max_ecc_key_length = 256;
+	card_caps.max_rsa_key_length = 2048;
+
+	if (card->version.fw_major >= 40) {
+	    /* Since 4.0, we can query available algorithms and key sizes.
+	     * Since 3.5.0 RSA up to 2048 and ECC up to 256 are always supported, so we check only max ECC key length. */
+	    if (myeid_get_card_caps(card, &card_caps) != SC_SUCCESS) {
+			sc_log(card->ctx, "Failed to get card capabilities. Using default max ECC key length 256.");
+	    }
+	}
+
+	flags = SC_ALGORITHM_RSA_RAW | SC_ALGORITHM_ONBOARD_KEY_GEN;
+	if (priv->disable_hw_pkcs1_padding == 0)
+		flags |= SC_ALGORITHM_RSA_PAD_PKCS1;
+	flags |= SC_ALGORITHM_RSA_HASH_NONE;
 
 	_sc_card_add_rsa_alg(card,  512, flags, 0);
 	_sc_card_add_rsa_alg(card,  768, flags, 0);
@@ -166,36 +264,69 @@ static int myeid_init(struct sc_card *card)
 	_sc_card_add_rsa_alg(card, 1536, flags, 0);
 	_sc_card_add_rsa_alg(card, 2048, flags, 0);
 
-	if (sc_hex_to_bin(atrp, defatr, &len) == 0
-		&& (len == card->atr.len) &&
-		memcmp(card->atr.value, defatr, len) == 0) {
-	    largeEccKeys = 1;
+	if (card_caps.card_supported_features & MYEID_CARD_CAP_RSA) {
+		if (card_caps.max_rsa_key_length >= 3072)
+			_sc_card_add_rsa_alg(card, 3072, flags, 0);
+		if (card_caps.max_rsa_key_length >= 4096)
+			_sc_card_add_rsa_alg(card, 4096, flags, 0);
 	}
 
 	/* show ECC algorithms if the applet version of the inserted card supports them */
-	if ((card->version.fw_major == 3 && card->version.fw_minor > 5) ||
-			card->version.fw_major >= 4)   {
+	if (card->version.fw_major >= 35) {
 		int i;
 
 		flags = SC_ALGORITHM_ECDSA_RAW | SC_ALGORITHM_ECDH_CDH_RAW | SC_ALGORITHM_ONBOARD_KEY_GEN;
-		flags |= SC_ALGORITHM_ECDSA_HASH_NONE | SC_ALGORITHM_ECDSA_HASH_SHA1;
+		flags |= SC_ALGORITHM_ECDSA_HASH_NONE;
 		ext_flags = SC_ALGORITHM_EXT_EC_NAMEDCURVE | SC_ALGORITHM_EXT_EC_UNCOMPRESES;
 
 		for (i=0; ec_curves[i].curve_name != NULL; i++) {
-		    if (i >= 2 && !largeEccKeys)
-			continue;
-		    else
-			_sc_card_add_ec_alg(card,  ec_curves[i].size, flags, ext_flags, &ec_curves[i].curve_oid);
+			if (card_caps.max_ecc_key_length >= ec_curves[i].size)
+				_sc_card_add_ec_alg(card, ec_curves[i].size, flags, ext_flags, &ec_curves[i].curve_oid);
 		}
 	}
 
+	/* show supported symmetric algorithms */
+	flags = 0;
+	if (card_caps.card_supported_features & MYEID_CARD_CAP_3DES) {
+		if (card_caps.max_des_key_length >= 64)
+			_sc_card_add_symmetric_alg(card, SC_ALGORITHM_DES, 64, flags);
+		if (card_caps.max_des_key_length >= 128)
+			_sc_card_add_symmetric_alg(card, SC_ALGORITHM_3DES, 128, flags);
+		if (card_caps.max_des_key_length >= 192)
+			_sc_card_add_symmetric_alg(card, SC_ALGORITHM_3DES, 192, flags);
+	}
+	if (card_caps.card_supported_features & MYEID_CARD_CAP_AES) {
+		if (card_caps.max_aes_key_length >= 128)
+			_sc_card_add_symmetric_alg(card, SC_ALGORITHM_AES, 128, flags);
+		if (card_caps.max_aes_key_length >= 256)
+			_sc_card_add_symmetric_alg(card, SC_ALGORITHM_AES, 256, flags);
+	}
+
 	/* State that we have an RNG */
 	card->caps |= SC_CARD_CAP_RNG | SC_CARD_CAP_ISO7816_PIN_INFO;
 
-	card->max_recv_size = 255;
+	if ((card->version.fw_major == 40 && card->version.fw_minor >= 10 )
+		|| card->version.fw_major >= 41)
+		card->caps |= SC_CARD_CAP_WRAP_KEY | SC_CARD_CAP_UNWRAP_KEY
+			   | SC_CARD_CAP_ONCARD_SESSION_OBJECTS;
+
+	if (card->version.fw_major >= 45)
+		priv->cap_chaining = 1;
+	if (card->version.fw_major >= 40)
+		card->max_recv_size = 256;
+	else
+		card->max_recv_size = 255;
 	card->max_send_size = 255;
 
-	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+	rv = SC_SUCCESS;
+
+err:
+	if (rv < 0) {
+		free(priv);
+		card->drv_data = old_drv_data;
+	}
+
+	LOG_FUNC_RETURN(card->ctx, rv);
 }
 
 static const struct sc_card_operations *iso_ops = NULL;
@@ -341,20 +472,18 @@ static int myeid_process_fci(struct sc_card *card, struct sc_file *file,
 		sc_log(card->ctx, "id (%X) sec_attr (%X %X %X)", file->id,
 			file->sec_attr[0],file->sec_attr[1],file->sec_attr[2]);
 	}
-	tag = sc_asn1_find_tag(NULL, buf, buflen, 0x8A, &taglen);
-	if (tag != NULL && taglen > 0)
-	{
-		if(tag[0] == MYEID_STATE_CREATION) {
-			file->status = SC_FILE_STATUS_CREATION;
-			sc_log(card->ctx, "File id (%X) status SC_FILE_STATUS_CREATION (0x%X)",
-					file->id, tag[0]);
-		}
-		else if(tag[0] == MYEID_STATE_ACTIVATED) {
-			file->status = SC_FILE_STATUS_ACTIVATED;
-			sc_log(card->ctx, "File id (%X) status SC_FILE_STATUS_ACTIVATED (0x%X)",
-					file->id, tag[0]);
-		}
-		priv->card_state = file->status;
+
+	priv->card_state = file->status;
+	switch (file->status) {
+		case SC_FILE_STATUS_CREATION:
+			file->acl_inactive = 1;
+			sc_log(card->ctx, "File id (%X) status SC_FILE_STATUS_CREATION", file->id);
+			break;
+		case SC_FILE_STATUS_ACTIVATED:
+			sc_log(card->ctx, "File id (%X) status SC_FILE_STATUS_ACTIVATED", file->id);
+			break;
+		default:
+			sc_log(card->ctx, "File id (%X) unusual status (0x%X)", file->id, file->status);
 	}
 
 	LOG_FUNC_RETURN(card->ctx, 0);
@@ -413,6 +542,8 @@ static int encode_file_structure(sc_card_t *card, const sc_file_t *file,
 	else   {
 		delete = sc_file_get_acl_entry(file, SC_AC_OP_DELETE);
 
+		sc_log(card->ctx, "id (%X), type (%X)", file->id, file->type);
+
 		switch (file->type) {
 		case SC_FILE_TYPE_WORKING_EF:
 
@@ -446,10 +577,13 @@ static int encode_file_structure(sc_card_t *card, const sc_file_t *file,
 	/* Proprietary Information */
 	buf[18] = 0x85;
 	buf[19] = 0x02;
-	/* AC right to clear default 0 */
-	/* TODO: Implement this */
-	buf[20] = 0x00; /*(SC_FILE_TYPE_INTERNAL_EF == file->type ? 0x00 : 0x80);*/
-	buf[21] = 0x00;
+	if (file->prop_attr_len == 2 && file->prop_attr != NULL)
+	    memcpy(&buf[20], file->prop_attr, 2);
+	else
+	{
+		buf[20] = 0x00;
+		buf[21] = 0x00;
+	}
 
 	/* Life Cycle Status tag */
 	buf[22] = 0x8A;
@@ -569,13 +703,15 @@ static int myeid_set_security_env_rsa(sc_card_t *card, const sc_security_env_t *
 	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
 	u8 *p;
 	int r;
+	size_t i;
+	sc_path_t *target_file;
 
 	assert(card != NULL && env != NULL);
 	LOG_FUNC_CALLED(card->ctx);
 
-	if (env->flags & SC_SEC_ENV_KEY_REF_ASYMMETRIC)
+	if (env->flags & SC_SEC_ENV_KEY_REF_SYMMETRIC)
 	{
-		sc_log(card->ctx, "asymmetric keyref not supported.\n");
+		sc_log(card->ctx, "symmetric keyref not supported.\n");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 	if (se_num > 0)
@@ -595,6 +731,14 @@ static int myeid_set_security_env_rsa(sc_card_t *card, const sc_security_env_t *
 		apdu.p1 = 0x41;
 		apdu.p2 = 0xB6;
 		break;
+	case SC_SEC_OPERATION_UNWRAP:
+		apdu.p1 = 0x41;
+		apdu.p2 = 0xB8;
+		break;
+	case SC_SEC_OPERATION_WRAP:
+		apdu.p1 = 0x81;
+		apdu.p2 = 0xB8;
+		break;
 	default:
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
@@ -613,12 +757,43 @@ static int myeid_set_security_env_rsa(sc_card_t *card, const sc_security_env_t *
 		memcpy(p, env->file_ref.value, 2);
 		p += 2;
 	}
-	if (env->flags & SC_SEC_ENV_KEY_REF_PRESENT)
+	if (env->flags & SC_SEC_ENV_KEY_REF_PRESENT && env->operation != SC_SEC_OPERATION_UNWRAP &&
+		env->operation != SC_SEC_OPERATION_WRAP)
 	{
 		*p++ = 0x84;
 		*p++ = 1;
 		*p++ = 0;
 	}
+	for (i = 0; i < SC_SEC_ENV_MAX_PARAMS; i++)
+	    if (env->params[i].param_type == SC_SEC_ENV_PARAM_TARGET_FILE) {
+			target_file = (sc_path_t*) env->params[i].value;
+			if (env->params[i].value_len < sizeof(sc_path_t) || target_file->len != 2) {
+				sc_log(card->ctx, "wrong length of target file reference.\n");
+				return SC_ERROR_WRONG_LENGTH;
+			}
+			*p++ = 0x83;
+			*p++ = 2;
+			memcpy(p, target_file->value, 2);
+			p+= 2;
+			break;
+	    }
+
+	if (env->operation ==  SC_SEC_OPERATION_UNWRAP || env->operation == SC_SEC_OPERATION_WRAP)
+	{
+	    /* add IV if present */
+		for (i = 0; i < SC_SEC_ENV_MAX_PARAMS; i++)
+			if (env->params[i].param_type == SC_SEC_ENV_PARAM_IV) {
+				*p++ = 0x87;
+				*p++ = (unsigned char) env->params[i].value_len;
+				if (p + env->params[i].value_len >= sbuf + SC_MAX_APDU_BUFFER_SIZE) {
+					sc_log(card->ctx, "IV too long.\n");
+					return SC_ERROR_WRONG_LENGTH;
+				}
+				memcpy(p, env->params[i].value, env->params[i].value_len);
+				p+=(unsigned char) env->params[i].value_len;
+				break;
+			}
+	}
 	r = p - sbuf;
 	apdu.lc = r;
 	apdu.datalen = r;
@@ -656,9 +831,9 @@ static int myeid_set_security_env_ec(sc_card_t *card, const sc_security_env_t *e
 	assert(card != NULL && env != NULL);
 	LOG_FUNC_CALLED(card->ctx);
 
-	if (env->flags & SC_SEC_ENV_KEY_REF_ASYMMETRIC)
+	if (env->flags & SC_SEC_ENV_KEY_REF_SYMMETRIC)
 	{
-		sc_log(card->ctx, "asymmetric keyref not supported.");
+		sc_log(card->ctx, "symmetric keyref not supported.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 	if (se_num > 0)
@@ -678,6 +853,10 @@ static int myeid_set_security_env_ec(sc_card_t *card, const sc_security_env_t *e
 		apdu.p1 = 0x41;
 		apdu.p2 = 0xB6;
 		break;
+	case SC_SEC_OPERATION_DERIVE:
+		apdu.p1 = 0x41;
+		apdu.p2 = 0xA4;
+		break;
 	default:
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
@@ -750,13 +929,20 @@ static int myeid_set_security_env(struct sc_card *card,
 
 		if (tmp.algorithm == SC_ALGORITHM_RSA)
 		{
-			tmp.algorithm_ref = 0x00;
-			/* potential FIXME: return an error, if an unsupported
-			* pad or hash was requested, although this shouldn't happen */
-			if (env->algorithm_flags & SC_ALGORITHM_RSA_PAD_PKCS1)
-				tmp.algorithm_ref = 0x02;
-			if (tmp.algorithm_flags & SC_ALGORITHM_RSA_HASH_SHA1)
-				tmp.algorithm_ref |= 0x10;
+			if (tmp.operation == SC_SEC_OPERATION_UNWRAP || tmp.operation == SC_SEC_OPERATION_WRAP)
+			{
+			    tmp.algorithm_ref = 0x0A;
+			}
+			else
+			{
+				tmp.algorithm_ref = 0x00;
+				/* potential FIXME: return an error, if an unsupported
+				* pad or hash was requested, although this shouldn't happen */
+				if (env->algorithm_flags & SC_ALGORITHM_RSA_PAD_PKCS1)
+					tmp.algorithm_ref = 0x02;
+				if (tmp.algorithm_flags & SC_ALGORITHM_RSA_HASH_SHA1)
+					tmp.algorithm_ref |= 0x10;
+			}
 
 			return myeid_set_security_env_rsa(card, &tmp, se_num);
 		}
@@ -766,6 +952,23 @@ static int myeid_set_security_env(struct sc_card *card,
 			tmp.algorithm_flags = 0;
 			return myeid_set_security_env_ec(card, &tmp, se_num);
 		}
+		else if (tmp.algorithm == SC_ALGORITHM_AES)
+		{
+			if (tmp.operation == SC_SEC_OPERATION_UNWRAP || tmp.operation == SC_SEC_OPERATION_WRAP)
+			{
+				tmp.algorithm_ref = 0x0A;
+			}
+			else
+			{
+				tmp.algorithm_ref = 0x00;
+			}
+
+			if ((tmp.algorithm_flags & SC_ALGORITHM_AES_CBC_PAD) == SC_ALGORITHM_AES_CBC_PAD)
+				tmp.algorithm_ref |= 0x80;		/* set PKCS#7 padding */
+
+			/* from this point, there's no difference to RSA SE */
+			return myeid_set_security_env_rsa(card, &tmp, se_num);
+		}
 		else
 		{
 
@@ -816,32 +1019,101 @@ myeid_convert_ec_signature(struct sc_context *ctx, size_t s_len, unsigned char *
 	if (sig_len != (datalen - len_size - 1))	/* validate size of the DER structure */
 	    return SC_ERROR_INVALID_DATA;
 
-	buf = calloc(1, (s_len + 7)/8*2);
-	if (!buf)
-		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+	/* test&fail early */
 	buflen = (s_len + 7)/8*2;
-
-	r = sc_asn1_sig_value_sequence_to_rs(ctx, data, datalen, buf, buflen);
-	if (r < 0)
-		free(buf);
-	LOG_TEST_RET(ctx, r, "Failed to cenvert Sig-Value to the raw RS format");
-
 	if (buflen > datalen)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_DATA);
 
+	buf = calloc(1, buflen);
+	if (!buf)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+
+	r = sc_asn1_sig_value_sequence_to_rs(ctx, data, datalen, buf, buflen);
+	if (r < 0) {
+		free(buf);
+		sc_log(ctx, "Failed to convert Sig-Value to the raw RS format");
+		return r;
+	}
+
 	memmove(data, buf, buflen);
+	free(buf);
 	return buflen;
 }
+/* MyEID cards before version 4.5 do not support RAW RSA signature for 2048 bit RSA keys.
+ * (Source: MyEID reference manual 2.1.4)
+ *
+ * This function uses decipher operation for calculating RAW 2048 bit signature. */
+static int
+myeid_compute_raw_2048_signature(struct sc_card *card, const u8 * data, size_t datalen,
+		u8 * out, size_t outlen)
+{
+	int r;
+	struct sc_context *ctx;
+	struct myeid_private_data *priv;
+	struct sc_apdu apdu;
+	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
+	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
+	sc_security_env_t env;
 
+	ctx = card->ctx;
+	LOG_FUNC_CALLED(ctx);
+
+	priv = (myeid_private_data_t *) card->drv_data;
+
+/* security env change - use DECIPHER operation */
+	memcpy(&env, priv->sec_env, sizeof(sc_security_env_t));
+	env.flags |= SC_SEC_ENV_ALG_REF_PRESENT;
+	env.flags |= SC_SEC_ENV_FILE_REF_PRESENT;
+	env.flags |= SC_SEC_ENV_KEY_REF_PRESENT;
+	env.operation = SC_SEC_OPERATION_DECIPHER;
+	myeid_set_security_env_rsa(card, &env, 0);
+
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x2A, 0x80, 0x86);
+	apdu.resp = rbuf;
+	apdu.resplen = sizeof(rbuf);
+	apdu.le = 0;	/* there is no response to 1st part of data */
+
+/* prepare 1st part of data */
+	sbuf[0] = 0x81;
+	memcpy(sbuf + 1, data, datalen / 2);
+	apdu.lc = datalen / 2 + 1;
+	apdu.datalen = apdu.lc;
+	apdu.data = sbuf;
+
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
+/* prepare 2nd part of data */
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x2A, 0x80, 0x86);
+		apdu.resp = rbuf;
+		apdu.resplen = sizeof(rbuf);
+		apdu.le = datalen;
+		sbuf[0] = 0x82;
+		memcpy(sbuf + 1, data + datalen / 2, datalen / 2);
+		apdu.lc = datalen / 2 + 1;
+		apdu.datalen = apdu.lc;
+		apdu.data = sbuf;
+
+		r = sc_transmit_apdu(card, &apdu);
+		LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+
+		if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
+			int len = apdu.resplen > outlen ? outlen : apdu.resplen;
+			memcpy(out, apdu.resp, len);
+			LOG_FUNC_RETURN(card->ctx, len);
+		}
+	}
+	LOG_FUNC_RETURN(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2));
+}
 
 static int
 myeid_compute_signature(struct sc_card *card, const u8 * data, size_t datalen,
 		u8 * out, size_t outlen)
 {
-	struct sc_context *ctx = card->ctx;
+	struct sc_context *ctx;
 	struct sc_apdu apdu;
-	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
-	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
+	u8 rbuf[MYEID_MAX_EXT_APDU_BUFFER_SIZE];
+	u8 sbuf[MYEID_MAX_EXT_APDU_BUFFER_SIZE];
 	struct myeid_private_data* priv;
 	int r;
 	size_t field_length = 0;
@@ -867,27 +1139,23 @@ myeid_compute_signature(struct sc_card *card, const u8 * data, size_t datalen,
 		}
 	}
 
-	if ((datalen + pad_chars) > 256)
+	if ((datalen + pad_chars) > sizeof(sbuf))
 		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
 
+	if (priv->sec_env->algorithm == SC_ALGORITHM_RSA && datalen == 256 && !priv->cap_chaining)
+		return myeid_compute_raw_2048_signature(card, data, datalen, out, outlen);
+
 	/* INS: 0x2A  PERFORM SECURITY OPERATION
 		* P1:  0x9E  Resp: Digital Signature
 		* P2:  0x9A  Cmd: Input for Digital Signature */
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x2A, 0x9E, 0x9A);
+	apdu.flags |= SC_APDU_FLAGS_CHAINING;
 	apdu.resp = rbuf;
 	apdu.resplen = sizeof(rbuf);
 	apdu.le = 256;
-	if (datalen == 256)   {
-		apdu.p2 = data[0];
-		memcpy(sbuf, data+1, datalen-1);
-		apdu.lc = datalen - 1;
-		apdu.datalen = datalen - 1;
-	}
-	else   {
-		memcpy(sbuf + pad_chars, data, datalen);
-		apdu.lc = datalen + pad_chars;
-		apdu.datalen = datalen + pad_chars;
-	}
+	memcpy(sbuf + pad_chars, data, datalen);
+	apdu.lc = datalen + pad_chars;
+	apdu.datalen = datalen + pad_chars;
 
 	apdu.data = sbuf;
 	r = sc_transmit_apdu(card, &apdu);
@@ -895,7 +1163,7 @@ myeid_compute_signature(struct sc_card *card, const u8 * data, size_t datalen,
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 	LOG_TEST_RET(ctx, r, "compute_signature failed");
 
-	if (priv->sec_env->algorithm == SC_ALGORITHM_EC)   {
+	if (priv->sec_env->algorithm == SC_ALGORITHM_EC) {
 		r = myeid_convert_ec_signature(ctx, priv->sec_env->algorithm_ref, apdu.resp, apdu.resplen);
 		LOG_TEST_RET(ctx, r, "compute_signature convert signature failed");
 		apdu.resplen = r;
@@ -920,20 +1188,43 @@ int myeid_ecdh_derive(struct sc_card *card, const u8* pubkey, size_t pubkey_len,
 	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
 
 	int r;
+	size_t ext_len_bytes;
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x86, 0x00, 0x00);
 
 	apdu.resp = rbuf;
 	apdu.resplen = sizeof(rbuf);
 
-	/* Fill in "Data objects in dynamic authentication template (tag 0x7C) structure */
-	sbuf[0] = 0x7C;
-	sbuf[1] = pubkey_len + 4;
-	sbuf[2] = 0x85;
-	sbuf[3] = pubkey_len;
-	memcpy(&sbuf[4], pubkey, pubkey_len);
+	/* Fill in "Data objects in dynamic authentication template" (tag 0x7C) structure
+	*
+	* TODO: encode the structure using OpenSC's ASN1-functions.
+	*
+	*  Size of the structure depends on key length. With 521 bit keys two bytes are needed for defining length of a point.
+	*/
 
-	apdu.lc = pubkey_len + 4;
+	sbuf[0] = 0x7C;
+	ext_len_bytes = 0;
+
+	if (pubkey_len > 127)
+	{
+		sbuf[1] = 0x81;
+		sbuf[2] = (u8) (pubkey_len + 3);
+		sbuf[3] = 0x85;
+		sbuf[4] = 0x81;
+		sbuf[5] = (u8) (pubkey_len);
+		ext_len_bytes = 2;
+	}
+	else
+	{
+		sbuf[1] = pubkey_len + 2;
+		sbuf[2] = 0x85;
+		sbuf[3] = pubkey_len;
+	}
+
+	memcpy(&sbuf[4 + ext_len_bytes], pubkey, pubkey_len);
+
+	apdu.lc = pubkey_len + 4 + ext_len_bytes;
+	apdu.le = pubkey_len / 2;
 	apdu.datalen = apdu.lc;
 	apdu.data = sbuf;
 
@@ -952,21 +1243,110 @@ int myeid_ecdh_derive(struct sc_card *card, const u8* pubkey, size_t pubkey_len,
 
 	memcpy(out, rbuf, apdu.resplen);
 
+	LOG_FUNC_RETURN(card->ctx, apdu.resplen);
+}
+
+static int myeid_transmit_decipher_pi_split(struct sc_card *card, struct sc_apdu *apdu, u8 *sbuf)
+{
+	/* MyEID before 4.5.x does not support APDU chaining. The payload
+	 * is split to two regular APDUs and Padding Indicator field is used to
+	 * describe which slice it is. */
+	size_t crgram_len = apdu->lc - 1;
+	size_t crgram_half = crgram_len / 2;
+	size_t resplen = apdu->resplen;
+	unsigned char *resp = apdu->resp;
+	int r;
+
+	LOG_FUNC_CALLED(card->ctx);
+
+	/* Send 1st part, no response */
+	apdu->cse = SC_APDU_CASE_3_SHORT;
+	apdu->data = &sbuf[0];
+	apdu->datalen = apdu->lc = crgram_half + 1;
+	apdu->resp = 0;
+	apdu->resplen = 0;
+	apdu->le = 0;
+	sbuf[0] = 0x81;			/* Padding Indicator, 0x81 = First half */
+
+	r = sc_transmit_apdu(card, apdu);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	if (apdu->sw1 != 0x90 || apdu->sw2 != 0x00)
+		return 0;
+
+	/* Send 2nd part, expect response */
+	apdu->cse = resplen ? SC_APDU_CASE_4_SHORT : SC_APDU_CASE_3_SHORT;
+	apdu->data = &sbuf[crgram_half];
+	apdu->datalen = apdu->lc = crgram_len - crgram_half + 1;
+	apdu->resp = resp;
+	apdu->resplen = resplen;
+	apdu->le = resplen ? MIN(card->max_recv_size, crgram_len) : 0;
+	sbuf[crgram_half] = 0x82;	/* Padding Indicator, 0x82 = Second half */
+
+	r = sc_transmit_apdu(card, apdu);
 	LOG_FUNC_RETURN(card->ctx, r);
 }
 
+static int myeid_transmit_decipher(struct sc_card *card, u8 p1, u8 p2,
+		const u8 * crgram, size_t crgram_len, u8 * out, size_t outlen)
+{
+	myeid_private_data_t *priv = card->drv_data;
+	struct sc_apdu apdu;
+	u8 rbuf[SC_MAX_EXT_APDU_BUFFER_SIZE];
+	u8 sbuf[SC_MAX_EXT_APDU_BUFFER_SIZE];
+	int r;
+
+	LOG_FUNC_CALLED(card->ctx);
+
+	/* INS: 0x2A  PERFORM SECURITY OPERATION
+	 * P1:  0x00  Resp: No response (unwrapping)
+	 * P1:  0x80  Resp: Plain value
+	 * P2:  0x84  Cmd: Cryptogram (no padding byte)
+	 * P2:  0x86  Cmd: Padding indicator byte followed by cryptogram */
+	sc_format_apdu(card, &apdu, p1 ? SC_APDU_CASE_4_SHORT : SC_APDU_CASE_3_SHORT, 0x2A, p1, p2);
+	if (p2 == 0x86) {
+		if (crgram_len+1 > sizeof(sbuf))
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+		sbuf[0] = 0; /* Padding indicator: 0x00 = No further indication */
+		memcpy(sbuf + 1, crgram, crgram_len);
+		apdu.data = sbuf;
+		apdu.datalen = apdu.lc = crgram_len + 1;
+	} else {
+		apdu.data = crgram;
+		apdu.datalen = apdu.lc = crgram_len;
+	}
+	if (p1 != 0x00) {
+		apdu.resp = rbuf;
+		apdu.resplen = sizeof(rbuf);
+		apdu.le = MIN(card->max_recv_size, crgram_len);
+	}
+
+	if (p2 == 0x86 && crgram_len == 256 && priv && !priv->cap_chaining) {
+		r = myeid_transmit_decipher_pi_split(card, &apdu, sbuf);
+	} else {
+		apdu.flags |= SC_APDU_FLAGS_CHAINING;
+		r = sc_transmit_apdu(card, &apdu);
+	}
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "DECIPHER returned error");
+
+	if (out && outlen) {
+		outlen = MIN(apdu.resplen, outlen);
+		memcpy(out, apdu.resp, outlen);
+	} else {
+		outlen = 0;
+	}
+	LOG_FUNC_RETURN(card->ctx, outlen);
+}
 
 static int myeid_decipher(struct sc_card *card, const u8 * crgram,
 		size_t crgram_len, u8 * out, size_t outlen)
 {
 	int r;
 	myeid_private_data_t* priv;
-	struct sc_apdu apdu;
-	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
-	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
 
 	LOG_FUNC_CALLED(card->ctx);
-	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_NORMAL);
 
 	assert(card != NULL && crgram != NULL && out != NULL);
 
@@ -974,84 +1354,81 @@ static int myeid_decipher(struct sc_card *card, const u8 * crgram,
 
 	if (priv->sec_env && priv->sec_env->algorithm == SC_ALGORITHM_EC
 		&& priv->sec_env->operation == SC_SEC_OPERATION_DERIVE
-		&& priv->sec_env->algorithm_flags & SC_ALGORITHM_ECDSA_RAW)
+		&& priv->sec_env->algorithm_flags & SC_ALGORITHM_ECDH_CDH_RAW)
 	{
 		r = myeid_ecdh_derive(card, crgram, crgram_len, out, outlen);
 		priv->sec_env = NULL; /* clear after operation */
 		LOG_FUNC_RETURN(card->ctx, r);
 	}
 
-	if (crgram_len > 256)
-		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+	r = myeid_transmit_decipher(card, 0x80, 0x86, crgram, crgram_len, out, outlen);
+	LOG_FUNC_RETURN(card->ctx, r);
+}
+
+
+static int myeid_wrap_key(struct sc_card *card, u8 *out, size_t outlen)
+{
+	struct sc_context *ctx;
+	struct sc_apdu apdu;
+	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
+	int r;
+
+	assert(card != NULL);
+	ctx = card->ctx;
+	LOG_FUNC_CALLED(ctx);
 
 	/* INS: 0x2A  PERFORM SECURITY OPERATION
-	    * P1:  0x80  Resp: Plain value
-	    * P2:  0x86  Cmd: Padding indicator byte followed by cryptogram */
-	sc_format_apdu(card, &apdu,
-		(crgram_len < 256) ? SC_APDU_CASE_4_SHORT : SC_APDU_CASE_3_SHORT,
-		0x2A, 0x80, 0x86);
-
+	   P1:  0x84  Resp: Return a cryptogram
+	 * P2:  0x00  The data field is absent */
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0x2A, 0x84, 0x00);
 	apdu.resp = rbuf;
 	apdu.resplen = sizeof(rbuf);
-	apdu.le = crgram_len;
-
-	if (crgram_len == 256)
-	{
-		apdu.le = 0;
-		/* padding indicator byte, 0x81 = first half of 2048 bit cryptogram */
-		sbuf[0] = 0x81;
-		memcpy(sbuf + 1, crgram, crgram_len / 2);
-		apdu.lc = crgram_len / 2 + 1;
-	}
-	else
-	{
-		sbuf[0] = 0; /* padding indicator byte, 0x00 = No further indication */
-		memcpy(sbuf + 1, crgram, crgram_len);
-		apdu.lc = crgram_len + 1;
-	}
-
-	apdu.datalen = apdu.lc;
-	apdu.data = sbuf;
+	apdu.le = 0;
+	apdu.lc = 0;
 
 	r = sc_transmit_apdu(card, &apdu);
-	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
-	if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00)
+	LOG_TEST_RET(ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(ctx, r, "wrap key failed");
+
+	if (apdu.resplen <= outlen && out != NULL)
+		memcpy(out, apdu.resp, apdu.resplen);
+
+	LOG_FUNC_RETURN(ctx, apdu.resplen);
+}
+
+static int myeid_unwrap_key(struct sc_card *card, const u8 *crgram, size_t crgram_len)
+{
+	myeid_private_data_t* priv;
+	u8 p2 = 0x86; /* init P2 for asymmetric crypto by default.*/
+	int r;
+
+	if (card == NULL || crgram == NULL)
+		return SC_ERROR_INVALID_ARGUMENTS;
+	priv = card->drv_data;
+
+	LOG_FUNC_CALLED(card->ctx);
+
+	if (crgram_len > MYEID_MAX_RSA_KEY_LEN / 8)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+
+	if (priv && priv->sec_env)
 	{
-		if (crgram_len == 256)
-		{
-			sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT,
-				0x2A, 0x80, 0x86);
-			apdu.resp = rbuf;
-			apdu.resplen = sizeof(rbuf);
-			apdu.le = crgram_len;
-			/* padding indicator byte,
-			    * 0x82 = Second half of 2048 bit cryptogram */
-			sbuf[0] = 0x82;
-			memcpy(sbuf + 1, crgram + crgram_len / 2, crgram_len / 2);
-			apdu.lc = crgram_len / 2 + 1;
-			apdu.datalen = apdu.lc;
-			apdu.data = sbuf;
-
-			r = sc_transmit_apdu(card, &apdu);
-
-			LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
-
-			if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00)
-			{
-				int len = apdu.resplen > outlen ? outlen : apdu.resplen;
-				memcpy(out, apdu.resp, len);
-				LOG_FUNC_RETURN(card->ctx, len);
-			}
-		}
-		else
-		{
-			int len = apdu.resplen > outlen ? outlen : apdu.resplen;
-
-			memcpy(out, apdu.resp, len);
-			LOG_FUNC_RETURN(card->ctx, len);
-		}
+		if (priv->sec_env->algorithm == SC_ALGORITHM_AES ||
+			priv->sec_env->algorithm == SC_ALGORITHM_3DES ||
+			priv->sec_env->algorithm == SC_ALGORITHM_DES)
+				p2 = 0x84;
 	}
-	LOG_FUNC_RETURN(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2));
+
+	if (p2 == 0x84 && crgram_len > MYEID_MAX_APDU_DATA_LEN)
+		LOG_TEST_RET(card->ctx, SC_ERROR_WRONG_LENGTH, "Unwrapping symmetric data longer that 255 bytes is not supported\n");
+
+	/* INS: 0x2A  PERFORM SECURITY OPERATION
+	 * P1:  0x00  Do not expect response - the deciphered data will be placed into the target key EF.
+	 * P2:  0x86  Cmd: Padding indicator byte followed by cryptogram
+	 * P2:  0x84  Cmd: AES/3DES Cryptogram (plain value encoded in BER-TLV DO, but not including SM DOs) */
+	r = myeid_transmit_decipher(card, 0x00, p2, crgram, crgram_len, 0, 0);
+	LOG_FUNC_RETURN(card->ctx, r);
 }
 
 
@@ -1118,77 +1495,55 @@ static int myeid_getdata(struct sc_card *card, struct sc_cardctl_myeid_data_obj*
 	LOG_FUNC_RETURN(card->ctx, r);
 }
 
-static int myeid_loadkey(sc_card_t *card, int mode, u8* value, int value_len)
+static int myeid_loadkey(sc_card_t *card, unsigned mode, u8* value, int value_len)
 {
+	myeid_private_data_t *priv = (myeid_private_data_t *) card->drv_data;
 	sc_apdu_t apdu;
-	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
-	int r, len;
+	u8 sbuf[MYEID_MAX_EXT_APDU_BUFFER_SIZE];
+	int r;
 
 	LOG_FUNC_CALLED(card->ctx);
-	len = 0;
-	if(value_len == 0 || value == NULL)
+	if (value_len == 0 || value == NULL)
 		return 0;
 
-	if(value    != NULL &&
-	   value[0] != 0x0 &&
-	   mode     != LOAD_KEY_PUBLIC_EXPONENT)
-		sbuf[len++] = 0x0;
-
-	if(mode == LOAD_KEY_MODULUS && value_len >= 256)
+	if (mode == LOAD_KEY_MODULUS && value_len == 256 && !priv->cap_chaining)
 	{
-		r=0;
-		if((value_len % 2) > 0 && value[0] == 0x00)
-		{
-			value_len--;
-			memmove(value, value + 1, value_len);
-		}
-		mode   = 0x88;
-		len    = 128;
-		memcpy(sbuf,value, 128);
-
+		mode = 0x88;
 		memset(&apdu, 0, sizeof(apdu));
 		sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xDA, 0x01, mode);
 
 		apdu.cla     = 0x00;
-		apdu.data    = sbuf;
-		apdu.datalen = len;
-		apdu.lc	     = len;
+		apdu.data    = value;
+		apdu.datalen = 128;
+		apdu.lc	     = 128;
 
 		r = sc_transmit_apdu(card, &apdu);
-		if(r < 0)
-			return r;
+		LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+
 		r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-		if(r < 0)
-			return r;
+		LOG_TEST_RET(card->ctx, r, "LOAD KEY returned error");
 
 		mode = 0x89;
-		len  = value_len - 128;
-		memset(&sbuf, 0, SC_MAX_APDU_BUFFER_SIZE);
-		memcpy(sbuf,value + 128, value_len - 128);
+		value += 128;
+		value_len -= 128;
 	}
-	else  if(mode == LOAD_KEY_EC_PRIVATE) {
-		memcpy(sbuf, value, value_len);
-		len = value_len;
-		mode = LOAD_KEY_MODE_EC_PRIV;
-	}
-	else if(mode == LOAD_KEY_EC_PUBLIC) {
-		memcpy(sbuf, value, value_len);
-		len = value_len;
-		mode = LOAD_KEY_MODE_EC_PUB;
-	}
-	else
+	else if ((mode & 0xff00) == 0 && mode != LOAD_KEY_PUBLIC_EXPONENT &&
+		 value[0] != 0x00)
 	{
-		memcpy(sbuf + len, value, value_len);
-		len += value_len;
+		/* RSA components needing leading zero byte */
+		sbuf[0] = 0x0;
+		memcpy(&sbuf[1], value, value_len);
+		value = sbuf;
+		value_len ++;
 	}
 
 	memset(&apdu, 0, sizeof(apdu));
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xDA, 0x01, mode);
-
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xDA, 0x01, mode & 0xFF);
+	apdu.flags   = SC_APDU_FLAGS_CHAINING;
 	apdu.cla     = 0x00;
-	apdu.data    = sbuf;
-	apdu.datalen = len;
-	apdu.lc	     = len;
+	apdu.data    = value;
+	apdu.datalen = value_len;
+	apdu.lc	     = value_len;
 
 	r = sc_transmit_apdu(card, &apdu);
 	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
@@ -1271,6 +1626,12 @@ static int myeid_generate_store_key(struct sc_card *card,
 					data->ecpublic_point_len)) >= 0)
 			LOG_FUNC_RETURN(card->ctx, r);
 		}
+		else if(data->key_type == SC_CARDCTL_MYEID_KEY_AES ||
+			data->key_type == SC_CARDCTL_MYEID_KEY_DES) {
+			if((r = myeid_loadkey(card, LOAD_KEY_SYMMETRIC, data->d,
+					data->d_len)) >= 0)
+			LOG_FUNC_RETURN(card->ctx, r);
+		}
 	}
 
 	LOG_FUNC_RETURN(card->ctx, r);
@@ -1360,6 +1721,45 @@ static int myeid_get_serialnr(sc_card_t *card, sc_serial_number_t *serial)
 	LOG_FUNC_RETURN(card->ctx, r);
 }
 
+/*
+ Get information of features that the card supports. MyEID 4.x cards are available on different
+ hardware and maximum key sizes cannot be determined simply from the version number anymore.
+ */
+static int myeid_get_card_caps(struct sc_card *card, myeid_card_caps_t* card_caps)
+{
+	sc_apdu_t apdu;
+	int r;
+	unsigned char rbuf[SC_MAX_APDU_BUFFER_SIZE];
+
+	LOG_FUNC_CALLED(card->ctx);
+
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xca, 0x01, 0xAA);
+	apdu.resp    = rbuf;
+	apdu.resplen = sizeof(myeid_card_caps_t);
+	apdu.le      = sizeof(myeid_card_caps_t);
+
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_RET(card->ctx, r,  "APDU transmit failed");
+
+	if (apdu.sw1 != 0x90 || apdu.sw2 != 0x00)
+		return SC_ERROR_INTERNAL;
+
+	if (apdu.resplen < 11) {
+		sc_log(card->ctx, "Unexpected response to GET DATA (MyEIC card capabilities)");
+		return SC_ERROR_INTERNAL;
+	}
+
+	card_caps->card_caps_ver = rbuf[0];
+	/* the card returns big endian values */
+	card_caps->card_supported_features = (unsigned short) rbuf[1] << 8 | rbuf[2];
+	card_caps->max_rsa_key_length = (unsigned short) rbuf[3] << 8 | rbuf[4];
+	card_caps->max_des_key_length = (unsigned short) rbuf[5] << 8 | rbuf[6];
+	card_caps->max_aes_key_length = (unsigned short) rbuf[7] << 8 | rbuf[8];
+	card_caps->max_ecc_key_length = (unsigned short) rbuf[9] << 8 | rbuf[10];
+
+	LOG_FUNC_RETURN(card->ctx, r);
+}
+
 static int myeid_card_ctl(struct sc_card *card, unsigned long cmd, void *ptr)
 {
 	int r = SC_ERROR_NOT_SUPPORTED;
@@ -1417,6 +1817,7 @@ static struct sc_card_driver * sc_get_driver(void)
 	myeid_ops.append_record		= NULL;
 	myeid_ops.update_record		= NULL;
 	myeid_ops.select_file		= myeid_select_file;
+	myeid_ops.get_response		= iso_ops->get_response;
 	myeid_ops.create_file		= myeid_create_file;
 	myeid_ops.delete_file		= myeid_delete_file;
 	myeid_ops.list_files		= myeid_list_files;
@@ -1426,6 +1827,8 @@ static struct sc_card_driver * sc_get_driver(void)
 	myeid_ops.process_fci		= myeid_process_fci;
 	myeid_ops.card_ctl		= myeid_card_ctl;
 	myeid_ops.pin_cmd		= myeid_pin_cmd;
+	myeid_ops.wrap			= myeid_wrap_key;
+	myeid_ops.unwrap		= myeid_unwrap_key;
 	return &myeid_drv;
 }
 
