@@ -205,6 +205,8 @@ enum {
 	SC_CARD_TYPE_IASECC_AMOS,
 	SC_CARD_TYPE_IASECC_MI,
 	SC_CARD_TYPE_IASECC_MI2,
+	SC_CARD_TYPE_IASECC_CPX,
+	SC_CARD_TYPE_IASECC_CPXCL,
 
 	/* SmartCard-HSM */
 	SC_CARD_TYPE_SC_HSM = 26000,
@@ -266,6 +268,8 @@ enum {
 	SC_CARD_TYPE_IDPRIME_BASE = 37000,
 	SC_CARD_TYPE_IDPRIME_V1,
 	SC_CARD_TYPE_IDPRIME_V2,
+	SC_CARD_TYPE_IDPRIME_V3,
+	SC_CARD_TYPE_IDPRIME_V4,
 	SC_CARD_TYPE_IDPRIME_GENERIC,
 
 	/* eDO cards */
