@@ -39,6 +39,11 @@ extern "C" {
 #include "libopensc/opensc.h"
 #include "libopensc/log.h"
 #include "libopensc/cards.h"
+#include "scconf/scconf.h"
+
+#ifdef ENABLE_OPENSSL
+#include "libopensc/sc-ossl-compat.h"
+#endif
 
 #define SC_FILE_MAGIC			0x14426950
 
@@ -107,6 +112,30 @@ unsigned long bebytes2ulong(const u8 *buf);
  */
 unsigned short bebytes2ushort(const u8 *buf);
 
+/**
+ * Convert 2 bytes in little endian order into an unsigned short
+ * @param  buf   the byte array of 2 bytes
+ * @return       the converted value
+ */
+unsigned short lebytes2ushort(const u8 *buf);
+/**
+ * Convert 4 bytes in little endian order into an unsigned long
+ * @param  buf   the byte array of 4 bytes
+ * @return       the converted value
+ */
+unsigned long lebytes2ulong(const u8 *buf);
+
+/* Usable for setting string elements of token info, which
+ * are either initialized to NULL or we need to clean
+ * previous value.
+ *
+ * @param   strp   The pointer where to store string
+ * @param   value  The string to store (is strdupped)
+ */
+void set_string(char **strp, const char *value);
+
+#define BYTES4BITS(num)  (((num) + 7) / 8)    /* number of bytes necessary to hold 'num' bits */
+
 /* Returns an scconf_block entry with matching ATR/ATRmask to the ATR specified,
  * NULL otherwise. Additionally, if card driver is not specified, search through
  * all card drivers user configured ATRs. */
@@ -114,14 +143,22 @@ scconf_block *_sc_match_atr_block(sc_context_t *ctx, struct sc_card_driver *driv
 
 /* Returns an index number if a match was found, -1 otherwise. table has to
  * be null terminated. */
-int _sc_match_atr(struct sc_card *card, struct sc_atr_table *table, int *type_out);
+int _sc_match_atr(struct sc_card *card, const struct sc_atr_table *table, int *type_out);
 
 int _sc_card_add_algorithm(struct sc_card *card, const struct sc_algorithm_info *info);
+int _sc_card_add_symmetric_alg(sc_card_t *card, unsigned int algorithm,
+			       unsigned int key_length, unsigned long flags);
 int _sc_card_add_rsa_alg(struct sc_card *card, unsigned int key_length,
-			 unsigned long flags, unsigned long exponent);
+		unsigned long flags, unsigned long exponent);
 int _sc_card_add_ec_alg(struct sc_card *card, unsigned int key_length,
-			 unsigned long flags, unsigned long ext_flags,
-			 struct sc_object_id *curve_oid);
+		unsigned long flags, unsigned long ext_flags,
+		struct sc_object_id *curve_oid);
+int _sc_card_add_eddsa_alg(struct sc_card *card, unsigned int key_length,
+		unsigned long flags, unsigned long ext_flags,
+		struct sc_object_id *curve_oid);
+int _sc_card_add_xeddsa_alg(struct sc_card *card, unsigned int key_length,
+		unsigned long flags, unsigned long ext_flags,
+		struct sc_object_id *curve_oid);
 
 /********************************************************************/
 /*                 pkcs1 padding/encoding functions                 */
@@ -142,11 +179,11 @@ int sc_pkcs1_strip_digest_info_prefix(unsigned int *algorithm,
  * @param  inlen   IN  length of the input
  * @param  out     OUT output buffer (in == out is allowed)
  * @param  outlen  OUT length of the output buffer
- * @param  modlen  IN  length of the modulus in bytes
+ * @param  mod_bits IN  length of the modulus in bits
  * @return SC_SUCCESS on success and an error code otherwise
  */
 int sc_pkcs1_encode(sc_context_t *ctx, unsigned long flags,
-	const u8 *in, size_t inlen, u8 *out, size_t *outlen, size_t modlen);
+		const u8 *in, size_t inlen, u8 *out, size_t *outlen, size_t mod_bits);
 /**
  * Get the necessary padding and sec. env. flags.
  * @param  ctx     IN  sc_contex_t object
@@ -157,8 +194,8 @@ int sc_pkcs1_encode(sc_context_t *ctx, unsigned long flags,
  * @return SC_SUCCESS on success and an error code otherwise
  */
 int sc_get_encoding_flags(sc_context_t *ctx,
-	unsigned long iflags, unsigned long caps,
-	unsigned long *pflags, unsigned long *salg);
+		unsigned long iflags, unsigned long caps,
+		unsigned long *pflags, unsigned long *salg);
 
 /********************************************************************/
 /*             mutex functions                                      */
@@ -220,7 +257,7 @@ unsigned long sc_thread_id(const sc_context_t *ctx);
  * @return SC_SUCCESS on success and an error code otherwise
  */
 int sc_apdu_get_octets(sc_context_t *ctx, const sc_apdu_t *apdu, u8 **buf,
-	size_t *len, unsigned int proto);
+		size_t *len, unsigned int proto);
 /**
  * Sets the status bytes and return data in the APDU
  * @param  ctx     sc_context_t object
@@ -230,22 +267,21 @@ int sc_apdu_get_octets(sc_context_t *ctx, const sc_apdu_t *apdu, u8 **buf,
  * @return SC_SUCCESS on success and an error code otherwise
  */
 int sc_apdu_set_resp(sc_context_t *ctx, sc_apdu_t *apdu, const u8 *buf,
-	size_t len);
+		size_t len);
 /**
  * Logs APDU
  * @param  ctx          sc_context_t object
- * @param  level	log if ctx->debug >= level
  * @param  buf          buffer with the APDU data
  * @param  len          length of the APDU
  * @param  is_outgoing  != 0 if the data is send to the card
  */
-#define sc_apdu_log(ctx, level, data, len, is_outgoing) \
-	sc_debug_hex(ctx, level, is_outgoing != 0 ? "Outgoing APDU" : "Incoming APDU", data, len)
+#define sc_apdu_log(ctx, data, len, is_outgoing) \
+	sc_debug_hex(ctx, SC_LOG_DEBUG_NORMAL, is_outgoing != 0 ? "Outgoing APDU" : "Incoming APDU", data, len)
 
 extern struct sc_reader_driver *sc_get_pcsc_driver(void);
 extern struct sc_reader_driver *sc_get_ctapi_driver(void);
 extern struct sc_reader_driver *sc_get_openct_driver(void);
-extern struct sc_reader_driver *sc_get_cardmod_driver(void);
+extern struct sc_reader_driver *sc_get_cryptotokenkit_driver(void);
 
 #ifdef __cplusplus
 }
