@@ -205,7 +205,7 @@ static int sc_pkcs15emu_gemsafeGPK_init(sc_pkcs15_card_t *p15card)
 
 	u8 sysrec[7];
 	int num_keyinfo = 0;
-	keyinfo kinfo[8]; /* will loook for 8 keys */
+	keyinfo kinfo[9]; /* will look for 9 keys */
 	u8 modulus_buf[ 1 + 1024 / 8]; /* tag+modulus */
 	u8 *cp;
 	char buf[256];
@@ -219,8 +219,8 @@ static int sc_pkcs15emu_gemsafeGPK_init(sc_pkcs15_card_t *p15card)
 
 	/* could read this off card if needed */
 
-	p15card->tokeninfo->label = strdup("GemSAFE");
-	p15card->tokeninfo->manufacturer_id = strdup(MANU_ID);
+	set_string(&p15card->tokeninfo->label, "GemSAFE");
+	set_string(&p15card->tokeninfo->manufacturer_id, MANU_ID);
 	/* get serial number */
 	r = sc_card_ctl(card, SC_CARDCTL_GET_SERIALNR, &serial);
 	if (r != SC_SUCCESS)
@@ -228,7 +228,7 @@ static int sc_pkcs15emu_gemsafeGPK_init(sc_pkcs15_card_t *p15card)
 	r = sc_bin_to_hex(serial.value, serial.len, buf, sizeof(buf), 0);
 	if (r != SC_SUCCESS)
 		return SC_ERROR_INTERNAL;
-	p15card->tokeninfo->serial_number = strdup(buf);
+	set_string(&p15card->tokeninfo->serial_number, buf);
 
 	/* test if we have a gemsafe app df */
 	memset(&path, 0, sizeof(path));
@@ -255,9 +255,9 @@ static int sc_pkcs15emu_gemsafeGPK_init(sc_pkcs15_card_t *p15card)
 
 	/* There may be more then one key in the directory. */
 	/* we need to find them so we can associate them with the */
-	/* the certificate.  The files are 0007 to 000f */
+	/* the certificate.  The files are 0007 to 000F */
 
-	for (i = 7; i < 16; i++) {
+	for (i = 0x7; i <= 0xF; i++) {
 		path.value[0] = 0x00;
 		path.value[1] = i;
 		path.len = 2;	
@@ -297,7 +297,7 @@ static int sc_pkcs15emu_gemsafeGPK_init(sc_pkcs15_card_t *p15card)
 		while (j--) 
 			*cp++ =  modulus_buf[j + 1];
 		num_keyinfo++;
-	} 
+	}
 
 	/* Get the gemsafe data with the cert */
 	 sc_format_path("3F000200004", &path);
@@ -360,14 +360,14 @@ static int sc_pkcs15emu_gemsafeGPK_init(sc_pkcs15_card_t *p15card)
 			}
 
 			if ( gsdata[idx1] == 0x30 &&
-				is_seq(gsdata + idx1, &seq_size1, &seq_len1) &&
-			 	is_seq(gsdata + idx1 + seq_size1, &seq_size2, &seq_len2) &&
-			    gsdata[idx1 + seq_size1 + seq_size2 + 0] == 0xa0 &&
-				gsdata[idx1 + seq_size1 + seq_size2 + 1] == 0x03 &&
-				gsdata[idx1 + seq_size1 + seq_size2 + 2] == 0x02 &&
-				gsdata[idx1 + seq_size1 + seq_size2 + 3] == 0x01 &&
-				gsdata[idx1 + seq_size1 + seq_size2 + 4] == 0x02 &&
-				idx1 + 4 + seq_len1 < file->size) {
+					is_seq(gsdata + idx1, &seq_size1, &seq_len1) &&
+					is_seq(gsdata + idx1 + seq_size1, &seq_size2, &seq_len2) &&
+					gsdata[idx1 + seq_size1 + seq_size2 + 0] == 0xa0 &&
+					gsdata[idx1 + seq_size1 + seq_size2 + 1] == 0x03 &&
+					gsdata[idx1 + seq_size1 + seq_size2 + 2] == 0x02 &&
+					gsdata[idx1 + seq_size1 + seq_size2 + 3] == 0x01 &&
+					gsdata[idx1 + seq_size1 + seq_size2 + 4] == 0x02 &&
+					idx1 + 4 + seq_len1 < file->size) {
 				/* we have a cert (I hope) */
 				/* read in rest if needed */
 				idxlen = idx1 + seq_len1 + 4 - idx2; 
@@ -386,7 +386,7 @@ static int sc_pkcs15emu_gemsafeGPK_init(sc_pkcs15_card_t *p15card)
 					return SC_ERROR_OUT_OF_MEMORY;
 
 				memcpy(cert_info.value.value, gsdata + idx1, cert_info.value.len);
-			idx1 = idx1 + cert_info.value.len;
+				idx1 = idx1 + cert_info.value.len;
 				break;
 			}
 			idx1++;
