@@ -70,6 +70,7 @@ static const pgp_pin_cfg_t	pin_cfg_v2[3] = {
 	{ "Admin PIN",      0x03, PGP_ADMIN_PIN_FLAGS, 8, 2 }
 };
 
+static struct sc_object_id curve25519_oid = {{1, 3, 6, 1, 4, 1, 3029, 1, 5, 1, -1}};
 
 #define PGP_SIG_PRKEY_USAGE	(SC_PKCS15_PRKEY_USAGE_SIGN \
 				| SC_PKCS15_PRKEY_USAGE_SIGNRECOVER \
@@ -169,6 +170,8 @@ sc_pkcs15emu_openpgp_init(sc_pkcs15_card_t *p15card)
 	sc_path_t path;
 	sc_file_t *file = NULL;
 
+	LOG_FUNC_CALLED(ctx);
+
 	set_string(&p15card->tokeninfo->label, "OpenPGP card");
 	set_string(&p15card->tokeninfo->manufacturer_id, "OpenPGP project");
 
@@ -203,6 +206,7 @@ sc_pkcs15emu_openpgp_init(sc_pkcs15_card_t *p15card)
 	 *  01-03:	max length of pins 1-3
 	 *  04-07:	tries left for pins 1-3
 	 */
+	sc_log(ctx, "Reading PW status bytes");
 	if ((r = read_file(card, "006E:0073:00C4", c4data, sizeof(c4data))) < 0)
 		goto failed;
 	if (r != 7) {
@@ -251,19 +255,22 @@ sc_pkcs15emu_openpgp_init(sc_pkcs15_card_t *p15card)
 	 *  20-39:	finger print for ENC key
 	 *  40-59:	finger print for AUT key
 	 */
+	sc_log(ctx, "Reading Fingerprints");
 	if ((r = read_file(card, "006E:0073:00C5", c5data, sizeof(c5data))) < 0)
 		goto failed;
 	if (r < 60) {
-		sc_log(ctx, 
+		sc_log(ctx,
 			"finger print bytes have unexpected length (expected 60, got %d)\n", r);
 		return SC_ERROR_OBJECT_NOT_VALID;
 	}
 
+	sc_log(ctx, "Adding private keys");
 	/* XXX: check if "halfkeys" can be stored with gpg2. If not, add key pairs in one loop */
 	for (i = 0; i < 3; i++) {
 		sc_pkcs15_prkey_info_t prkey_info;
 		sc_pkcs15_object_t     prkey_obj;
 		u8 cxdata[12];
+		int cxdata_len = sizeof(cxdata);
 		char path_template[] = "006E:0073:00Cx";
 		int j;
 
@@ -272,7 +279,7 @@ sc_pkcs15emu_openpgp_init(sc_pkcs15_card_t *p15card)
 		memset(&cxdata, 0, sizeof(cxdata));
 
 		path_template[13] = '1' + i; /* The needed tags are C1 C2 and C3 */
-		if ((r = read_file(card, path_template, cxdata, sizeof(cxdata))) < 0)
+		if ((cxdata_len = read_file(card, path_template, cxdata, sizeof(cxdata))) < 1)
 			goto failed;
 
 		/* check validity using finger prints */
@@ -283,6 +290,10 @@ sc_pkcs15emu_openpgp_init(sc_pkcs15_card_t *p15card)
 
 		/* only add valid keys, i.e. those with a legal algorithm identifier & finger print */
 		if (j >= 0 && cxdata[0] != 0) {
+			struct sc_object_id oid;
+			struct sc_algorithm_info * algorithm_info; /* no need to free */
+
+			algorithm_info = NULL;
 			prkey_info.id.len         = 1;
 			prkey_info.id.value[0]    = i + 1;
 			prkey_info.usage          = key_cfg[i].prkey_usage;
@@ -294,24 +305,101 @@ sc_pkcs15emu_openpgp_init(sc_pkcs15_card_t *p15card)
 			prkey_obj.auth_id.len      = 1;
 			prkey_obj.auth_id.value[0] = key_cfg[i].prkey_pin;
 
-			if (cxdata[0] == SC_OPENPGP_KEYALGO_RSA && r >= 3) {
-				prkey_info.modulus_length = bebytes2ushort(cxdata + 1);
-				r = sc_pkcs15emu_add_rsa_prkey(p15card, &prkey_obj, &prkey_info);
+			/* need to get size from algorithms using oid */
+			if (cxdata[0] == SC_OPENPGP_KEYALGO_ECDH ||
+				cxdata[0] == SC_OPENPGP_KEYALGO_ECDSA ||
+				cxdata[0] == SC_OPENPGP_KEYALGO_EDDSA) {
+				/* Last byte could be Import-Format of private key, let's ignore it,
+				 * as it is not part of OID */
+				if (cxdata[cxdata_len-1] == SC_OPENPGP_KEYFORMAT_EC_STD ||
+				    cxdata[cxdata_len-1] == SC_OPENPGP_KEYFORMAT_EC_STDPUB)
+					cxdata_len--;
+				r = sc_asn1_decode_object_id(&cxdata[1], cxdata_len-1, &oid);
+				if (r != SC_SUCCESS) {
+					sc_log(ctx, "Failed to parse OID for elliptic curve algorithm");
+				}
 			}
-			if (cxdata[0] == SC_OPENPGP_KEYALGO_ECDH
-			   || cxdata[0] == SC_OPENPGP_KEYALGO_ECDSA) {
+
+			switch (cxdata[0]) {
+			case SC_OPENPGP_KEYALGO_ECDH:
+				if (sc_compare_oid(&oid, &curve25519_oid)) {
+					if ((algorithm_info = sc_card_find_xeddsa_alg(card, 0, &oid)))
+						prkey_info.field_length = algorithm_info->key_length;
+					else {
+						sc_log(ctx, "algorithm not found");
+						continue;
+					}
+					break;
+				}
+				/* Fall through */
+			case SC_OPENPGP_KEYALGO_ECDSA:
+				if((algorithm_info = sc_card_find_ec_alg(card, 0, &oid)))
+					prkey_info.field_length = algorithm_info->key_length;
+				else {
+					sc_log(ctx, "algorithm not found");
+					continue;
+				}
+				break;
+			case SC_OPENPGP_KEYALGO_EDDSA:
+				if ((algorithm_info = sc_card_find_eddsa_alg(card, 0, &oid)))
+					prkey_info.field_length = algorithm_info->key_length;
+				else {
+					sc_log(ctx, "algorithm not found");
+					continue;
+				}
+				break;
+			}
+
+			switch (cxdata[0]) {
+			case SC_OPENPGP_KEYALGO_EDDSA:
+				/* Filter out invalid usage: EdDSA does not support anything but sign */
+				prkey_info.usage &= PGP_SIG_PRKEY_USAGE;
+				r = sc_pkcs15emu_add_eddsa_prkey(p15card, &prkey_obj, &prkey_info);
+				break;
+
+			case SC_OPENPGP_KEYALGO_ECDH:
+				/* This can result in either ECDSA key or EC_MONTGOMERY
+				 * so we need to check OID */
+				if (sc_compare_oid(&oid, &curve25519_oid)) {
+					/* This can do only DERIVE */
+					prkey_info.usage = SC_PKCS15_PRKEY_USAGE_DERIVE;
+					r = sc_pkcs15emu_add_xeddsa_prkey(p15card, &prkey_obj, &prkey_info);
+					break;
+				}
+				prkey_info.usage |= SC_PKCS15_PRKEY_USAGE_DERIVE;
+				prkey_info.usage &= ~PGP_ENC_PRKEY_USAGE;
 				r = sc_pkcs15emu_add_ec_prkey(p15card, &prkey_obj, &prkey_info);
+				break;
+
+			case SC_OPENPGP_KEYALGO_ECDSA:
+				prkey_info.usage = SC_PKCS15_PRKEY_USAGE_SIGN;
+				r = sc_pkcs15emu_add_ec_prkey(p15card, &prkey_obj, &prkey_info);
+				break;
+
+			case SC_OPENPGP_KEYALGO_RSA:
+				if (cxdata_len >= 3) {
+					prkey_info.modulus_length = bebytes2ushort(cxdata + 1);
+					r = sc_pkcs15emu_add_rsa_prkey(p15card, &prkey_obj, &prkey_info);
+					break;
+				}
+				/* Fallthrough */
+			default:
+				sc_log(ctx, "Invalid algorithm identifier %x (length = %d)",
+					cxdata[0], r);
 			}
 
 			if (r < 0)
 				return SC_ERROR_INTERNAL;
 		}
 	}
+
+	sc_log(ctx, "Adding public keys");
 	/* Add public keys */
 	for (i = 0; i < 3; i++) {
 		sc_pkcs15_pubkey_info_t pubkey_info;
 		sc_pkcs15_object_t      pubkey_obj;
-		u8 cxdata[10];
+		u8 cxdata[12];
+		int cxdata_len = sizeof(cxdata);
 		char path_template[] = "006E:0073:00Cx";
 		int j;
 
@@ -320,7 +408,7 @@ sc_pkcs15emu_openpgp_init(sc_pkcs15_card_t *p15card)
 		memset(&cxdata, 0, sizeof(cxdata));
 
 		path_template[13] = '1' + i; /* The needed tags are C1 C2 and C3 */
-		if ((r = read_file(card, path_template, cxdata, sizeof(cxdata))) < 0)
+		if ((cxdata_len = read_file(card, path_template, cxdata, sizeof(cxdata))) < 1)
 			goto failed;
 
 		/* check validity using finger prints */
@@ -331,6 +419,10 @@ sc_pkcs15emu_openpgp_init(sc_pkcs15_card_t *p15card)
 
 		/* only add valid keys, i.e. those with a legal algorithm identifier & finger print */
 		if (j >= 0 && cxdata[0] != 0) {
+			struct sc_object_id oid;
+			struct sc_algorithm_info * algorithm_info; /* no need to free */
+
+			algorithm_info = NULL;
 			pubkey_info.id.len         = 1;
 			pubkey_info.id.value[0]    = i + 1;
 			pubkey_info.usage          = key_cfg[i].pubkey_usage;
@@ -339,13 +431,83 @@ sc_pkcs15emu_openpgp_init(sc_pkcs15_card_t *p15card)
 			strlcpy(pubkey_obj.label, key_cfg[i].label, sizeof(pubkey_obj.label));
 			pubkey_obj.flags = SC_PKCS15_CO_FLAG_MODIFIABLE;
 
-			if (cxdata[0] == SC_OPENPGP_KEYALGO_RSA && r >= 3) {
-				pubkey_info.modulus_length = bebytes2ushort(cxdata + 1);
-				r = sc_pkcs15emu_add_rsa_pubkey(p15card, &pubkey_obj, &pubkey_info);
+			if (cxdata[0] == SC_OPENPGP_KEYALGO_ECDH ||
+				cxdata[0] == SC_OPENPGP_KEYALGO_ECDSA ||
+				cxdata[0] == SC_OPENPGP_KEYALGO_EDDSA) {
+				/* Last byte could be Import-Format of private key, let's ignore it,
+				 * as it is not part of OID */
+				if (cxdata[cxdata_len-1] == SC_OPENPGP_KEYFORMAT_EC_STD ||
+				    cxdata[cxdata_len-1] == SC_OPENPGP_KEYFORMAT_EC_STDPUB)
+					cxdata_len--;
+				r = sc_asn1_decode_object_id(&cxdata[1], cxdata_len-1, &oid);
+				if (r != SC_SUCCESS) {
+					sc_log(ctx, "Failed to parse OID for elliptic curve algorithm");
+				}
 			}
-			if (cxdata[0] == SC_OPENPGP_KEYALGO_ECDH
-			   || cxdata[0] == SC_OPENPGP_KEYALGO_ECDSA) {
+
+			switch (cxdata[0]) {
+			case SC_OPENPGP_KEYALGO_ECDH:
+				if (sc_compare_oid(&oid, &curve25519_oid)) {
+					if ((algorithm_info = sc_card_find_xeddsa_alg(card, 0, &oid)))
+						pubkey_info.field_length = algorithm_info->key_length;
+					else {
+						sc_log(ctx, "algorithm not found");
+						continue;
+					}
+					break;
+				}
+				/* Fall through */
+			case SC_OPENPGP_KEYALGO_ECDSA:
+				if((algorithm_info = sc_card_find_ec_alg(card, 0, &oid)))
+					pubkey_info.field_length = algorithm_info->key_length;
+				else {
+					sc_log(ctx, "algorithm not found");
+					continue;
+				}
+				break;
+			case SC_OPENPGP_KEYALGO_EDDSA:
+				if ((algorithm_info = sc_card_find_eddsa_alg(card, 0, &oid)))
+					pubkey_info.field_length = algorithm_info->key_length;
+				else {
+					sc_log(ctx, "algorithm not found");
+					continue;
+				}
+				break;
+			}
+
+			switch (cxdata[0]) {
+			case SC_OPENPGP_KEYALGO_EDDSA:
+				/* assuming Ed25519 as it is the only supported now */
+				/* Filter out invalid usage: ED does not support anything but sign */
+				pubkey_info.usage &= PGP_SIG_PUBKEY_USAGE;
+				r = sc_pkcs15emu_add_eddsa_pubkey(p15card, &pubkey_obj, &pubkey_info);
+				break;
+			case SC_OPENPGP_KEYALGO_ECDH:
+				/* This can result in either ECDSA key or EC_MONTGOMERY
+				 * so we need to check OID */
+				if (sc_compare_oid(&oid, &curve25519_oid)) {
+					/* XXX What can this key do? */
+					pubkey_info.usage = SC_PKCS15_PRKEY_USAGE_DERIVE;
+					r = sc_pkcs15emu_add_xeddsa_pubkey(p15card, &pubkey_obj, &pubkey_info);
+					break;
+				}
+				pubkey_info.usage = SC_PKCS15_PRKEY_USAGE_DERIVE;
 				r = sc_pkcs15emu_add_ec_pubkey(p15card, &pubkey_obj, &pubkey_info);
+				break;
+			case SC_OPENPGP_KEYALGO_ECDSA:
+				pubkey_info.usage = PGP_SIG_PUBKEY_USAGE;
+				r = sc_pkcs15emu_add_ec_pubkey(p15card, &pubkey_obj, &pubkey_info);
+				break;
+			case SC_OPENPGP_KEYALGO_RSA:
+				if (cxdata_len >= 3) {
+					pubkey_info.modulus_length = bebytes2ushort(cxdata + 1);
+					r = sc_pkcs15emu_add_rsa_pubkey(p15card, &pubkey_obj, &pubkey_info);
+					break;
+				}
+				/* Fall through */
+			default:
+				sc_log(ctx, "Invalid algorithm identifier %x (length = %d)",
+					cxdata[0], r);
 			}
 
 			if (r < 0)
@@ -388,13 +550,13 @@ sc_pkcs15emu_openpgp_init(sc_pkcs15_card_t *p15card)
 
 failed:
 	if (r < 0) {
-		sc_log(card->ctx, 
+		sc_log(card->ctx,
 				"Failed to initialize OpenPGP emulation: %s\n",
 				sc_strerror(r));
 	}
 	sc_file_free(file);
 
-	return r;
+	LOG_FUNC_RETURN(ctx, r);
 }
 
 static int
