@@ -527,7 +527,9 @@ sc_pkcs15_prkey_attrs_from_cert(struct sc_pkcs15_card *p15card, struct sc_pkcs15
 
 	key_info = (struct sc_pkcs15_prkey_info *) key_object->data;
 
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
 	ERR_load_ERR_strings();
+#endif
 	ERR_load_crypto_strings();
 
 	sc_log(ctx, "CertValue(%"SC_FORMAT_LEN_SIZE_T"u) %p",
@@ -617,6 +619,14 @@ sc_pkcs15_free_prkey(struct sc_pkcs15_prkey *key)
 		if (key->u.ec.ecpointQ.value)
 			free(key->u.ec.ecpointQ.value);
 		break;
+	case SC_ALGORITHM_EDDSA:
+		free(key->u.eddsa.pubkey.value);
+		key->u.eddsa.pubkey.value = NULL;
+		key->u.eddsa.pubkey.len = 0;
+		free(key->u.eddsa.value.value);
+		key->u.eddsa.value.value = NULL;
+		key->u.eddsa.value.len = 0;
+		break;
 	}
 }
 
@@ -718,13 +728,13 @@ sc_pkcs15_convert_prkey(struct sc_pkcs15_prkey *pkcs15_key, void *evp_key)
 		}
 	case EVP_PKEY_EC: {
 		struct sc_pkcs15_prkey_ec *dst = &pkcs15_key->u.ec;
-		EC_KEY *src = NULL;
+		const EC_KEY *src = NULL;
 		const EC_GROUP *grp = NULL;
 		unsigned char buf[255];
 		size_t buflen = 255;
 		int nid;
 
-		src = EVP_PKEY_get0(pk);
+		src = EVP_PKEY_get0_EC_KEY(pk);
 		assert(src);
 		assert(EC_KEY_get0_private_key(src));
 		assert(EC_KEY_get0_public_key(src));
@@ -783,6 +793,13 @@ sc_pkcs15_convert_prkey(struct sc_pkcs15_prkey *pkcs15_key, void *evp_key)
 		break;
 	}
 #endif /* !defined(OPENSSL_NO_EC) */
+#ifdef EVP_PKEY_ED25519
+	case EVP_PKEY_ED25519: {
+		/* TODO */
+		break;
+	}
+#endif /* EVP_PKEY_ED25519 */
+
 	default:
 		return SC_ERROR_NOT_SUPPORTED;
 	}
