@@ -18,11 +18,14 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <stdio.h>
 #include <string.h>
 #include <assert.h>
+#include <ctype.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -34,8 +37,10 @@ int sc_decipher(sc_card_t *card,
 {
 	int r;
 
-	assert(card != NULL && crgram != NULL && out != NULL);
-	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_NORMAL);
+	if (card == NULL || crgram == NULL || out == NULL) {
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+	}
+	LOG_FUNC_CALLED(card->ctx);
 	if (card->ops->decipher == NULL)
 		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->decipher(card, crgram, crgram_len, out, outlen);
@@ -48,22 +53,56 @@ int sc_compute_signature(sc_card_t *card,
 {
 	int r;
 
-	assert(card != NULL);
-	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_NORMAL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	LOG_FUNC_CALLED(card->ctx);
 	if (card->ops->compute_signature == NULL)
 		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->compute_signature(card, data, datalen, out, outlen);
         SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, r);
 }
 
+int sc_unwrap(sc_card_t *card,
+		const u8 * crgram, size_t crgram_len, u8 * out, size_t outlen)
+{
+	int r;
+
+	if (card == NULL || crgram == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	LOG_FUNC_CALLED(card->ctx);
+	if (card->ops->unwrap == NULL)
+		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_NOT_SUPPORTED);
+	r = card->ops->unwrap(card, crgram, crgram_len);
+	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, r);
+}
+
+int sc_wrap(sc_card_t *card,
+		const u8 * crgram, size_t crgram_len, u8 * out, size_t outlen)
+{
+	int r;
+
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	LOG_FUNC_CALLED(card->ctx);
+	if (card->ops->wrap == NULL)
+		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_NOT_SUPPORTED);
+	r = card->ops->wrap(card, out, outlen);
+	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, r);
+}
+
 int sc_set_security_env(sc_card_t *card,
 			const sc_security_env_t *env,
 			int se_num)
 {
 	int r;
 
-	assert(card != NULL);
-	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_NORMAL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	LOG_FUNC_CALLED(card->ctx);
 	if (card->ops->set_security_env == NULL)
 		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->set_security_env(card, env, se_num);
@@ -74,8 +113,10 @@ int sc_restore_security_env(sc_card_t *card, int se_num)
 {
 	int r;
 
-	assert(card != NULL);
-	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_NORMAL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	LOG_FUNC_CALLED(card->ctx);
 	if (card->ops->restore_security_env == NULL)
 		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_NOT_SUPPORTED);
 	r = card->ops->restore_security_env(card, se_num);
@@ -151,10 +192,19 @@ int sc_reset_retry_counter(sc_card_t *card, unsigned int type, int ref,
 int sc_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 		int *tries_left)
 {
-	int r;
+	int r, debug;
+
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	LOG_FUNC_CALLED(card->ctx);
+
+	debug = card->ctx->debug;
+	if (data->cmd != SC_PIN_CMD_GET_INFO
+			&& card->ctx->debug < SC_LOG_DEBUG_PIN) {
+		card->ctx->debug = 0;
+	}
 
-	assert(card != NULL);
-	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_NORMAL);
 	if (card->ops->pin_cmd) {
 		r = card->ops->pin_cmd(card, data, tries_left);
 	} else if (!(data->flags & SC_PIN_CMD_USE_PINPAD)) {
@@ -195,12 +245,14 @@ int sc_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 			break;
 		}
 		if (r == SC_ERROR_NOT_SUPPORTED)
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "unsupported PIN operation (%d)",
+			sc_log(card->ctx,  "unsupported PIN operation (%d)",
 					data->cmd);
 	} else {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Use of pin pad not supported by card driver");
+		sc_log(card->ctx,  "Use of pin pad not supported by card driver");
 		r = SC_ERROR_NOT_SUPPORTED;
 	}
+	card->ctx->debug = debug;
+
 	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, r);
 }
 
@@ -208,10 +260,10 @@ int sc_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
  * This function will copy a PIN, convert and pad it as required
  *
  * Note about the SC_PIN_ENCODING_GLP encoding:
- * PIN buffers are allways 16 nibbles (8 bytes) and look like this:
+ * PIN buffers are always 16 nibbles (8 bytes) and look like this:
  *   0x2 + len + pin_in_BCD + paddingnibbles
  * in which the paddingnibble = 0xF
- * E.g. if PIN = 12345, then sbuf = {0x24, 0x12, 0x34, 0x5F, 0xFF, 0xFF, 0xFF, 0xFF}
+ * E.g. if PIN = 12345, then sbuf = {0x25, 0x12, 0x34, 0x5F, 0xFF, 0xFF, 0xFF, 0xFF}
  * E.g. if PIN = 123456789012, then sbuf = {0x2C, 0x12, 0x34, 0x56, 0x78, 0x90, 0x12, 0xFF}
  * Reference: Global Platform - Card Specification - version 2.0.1' - April 7, 2000
  */
@@ -231,7 +283,7 @@ int sc_build_pin(u8 *buf, size_t buflen, struct sc_pin_cmd_pin *pin, int pad)
 			if (pin->data[i] < '0' || pin->data[i] > '9')
 				return SC_ERROR_INVALID_ARGUMENTS;
 		}
-		buf[0] = 0x20 | pin_len;
+		buf[0] = 0x20 | (u8) pin_len;
 		buf++;
 		buflen--;
 	}
@@ -246,6 +298,9 @@ int sc_build_pin(u8 *buf, size_t buflen, struct sc_pin_cmd_pin *pin, int pad)
 		if (pin_len > 2 * buflen)
 			return SC_ERROR_BUFFER_TOO_SMALL;
 		for (i = j = 0; j < pin_len; j++) {
+			if (!isdigit(pin->data[j])) {
+				return SC_ERROR_INVALID_DATA;
+			}
 			buf[i] <<= 4;
 			buf[i] |= pin->data[j] & 0xf;
 			if (j & 1)
