@@ -225,10 +225,27 @@ typedef struct sc_acl_entry {
 #define SC_FILE_EF_CYCLIC_TLV		0x07
 
 /* File status flags */
-#define SC_FILE_STATUS_ACTIVATED	0x00
-#define SC_FILE_STATUS_INVALIDATED	0x01
-#define SC_FILE_STATUS_CREATION		0x02 /* Full access in this state,
-						(at least for SetCOS 4.4 */
+/* ISO7816-4: Unless otherwise specified, the security attributes are valid for the operational state.*/
+#define SC_FILE_STATUS_ACTIVATED	0x00 /* ISO7816-4: Operational state (activated)   (5, 7) */
+#define SC_FILE_STATUS_INVALIDATED	0x01 /* ISO7816-4: Operational state (deactivated) (4, 6) */
+
+/* Full access in this state, (at least for SetCOS 4.4 ) */
+#define SC_FILE_STATUS_CREATION		0x02 /* ISO7816-4: Creation state, (1) */
+
+#define SC_FILE_STATUS_INITIALISATION	0x03 /* ISO7816-4: Initialisation state, (3) */
+#define SC_FILE_STATUS_NO_INFO		0x04 /* ISO7816-4: No information given, (0) */
+#define SC_FILE_STATUS_TERMINATION	0x0c /* ISO7816-4: Termination state (12,13,14,15) */
+#define SC_FILE_STATUS_PROPRIETARY	0xf0 /* ISO7816-4: codes > 15 */
+
+/* reserved for future use by ISO/IEC */
+#define SC_FILE_STATUS_RFU_2		0x07 /* ISO7816-4: (0x02) */
+#define SC_FILE_STATUS_RFU_8		0x08 /* ISO7816-4: (0x08) */
+#define SC_FILE_STATUS_RFU_9		0x09 /* ISO7816-4: (0x09) */
+#define SC_FILE_STATUS_RFU_10		0x0a /* ISO7816-4: (0x0a) */
+#define SC_FILE_STATUS_RFU_11		0x0b /* ISO7816-4: (0x0b) */
+
+#define SC_FILE_STATUS_UNKNOWN		0xff /* if tag 0x8A is missing, there is no information about LCSB */
+
 typedef struct sc_file {
 	struct sc_path path;
 	unsigned char name[16];	/* DF name */
@@ -240,6 +257,7 @@ typedef struct sc_file {
 	int id;		/* file identifier (2 bytes) */
 	int sid;	/* short EF identifier (1 byte) */
 	struct sc_acl_entry *acl[SC_MAX_AC_OPS]; /* Access Control List */
+	int acl_inactive;                        /* if set, the card access control mechanism is not active */
 
 	size_t record_length; /* max. length in case of record-oriented EF */
 	size_t record_count;  /* Valid, if not transparent EF or DF */
