@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003 Mathias Brossard <mathias.brossard@idealx.com>
+ * Copyright (C) 2015 Mathias Brossard <mathias@brossard.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -21,41 +21,172 @@
 
 #include <stdlib.h>
 #include <stdio.h>
+#include <string.h>
 
 #ifdef _WIN32
 #include <windows.h>
 #include <winreg.h>
 #include <limits.h>
 #else
+#ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
+#endif
 #include <time.h>
 #endif
 
 #define CRYPTOKI_EXPORTS
 #include "pkcs11-display.h"
+#include "common/libpkcs11.h"
 
 #define __PASTE(x,y)      x##y
 
-extern void *C_LoadModule(const char *name, CK_FUNCTION_LIST_PTR_PTR);
-extern CK_RV C_UnloadModule(void *module);
-
 /* Declare all spy_* Cryptoki function */
 
 /* Spy Module Function List */
 static CK_FUNCTION_LIST_PTR pkcs11_spy = NULL;
+static CK_FUNCTION_LIST_3_0_PTR pkcs11_spy_3_0 = NULL;
 /* Real Module Function List */
-static CK_FUNCTION_LIST_PTR po = NULL;
+static CK_FUNCTION_LIST_3_0_PTR po = NULL;
 /* Dynamic Module Handle */
 static void *modhandle = NULL;
 /* Spy module output */
 static FILE *spy_output = NULL;
 
-/* Inits the spy. If successfull, po != NULL */
+static void *
+allocate_function_list(int v3)
+{
+	CK_FUNCTION_LIST_PTR list = NULL;
+	CK_FUNCTION_LIST_3_0_PTR list_3_0 = NULL;
+
+	if (v3) {
+		list = malloc(sizeof(CK_FUNCTION_LIST_3_0));
+	} else {
+		list = malloc(sizeof(CK_FUNCTION_LIST));
+	}
+	if (list == NULL) {
+		return NULL;
+	}
+	/* with our own pkcs11.h we need to maintain this ourself */
+	list->version.major = 2;
+	list->version.minor = 11;
+	list->C_Initialize = C_Initialize;
+	list->C_Finalize = C_Finalize;
+	list->C_GetInfo = C_GetInfo;
+	list->C_GetFunctionList = C_GetFunctionList;
+	list->C_GetSlotList = C_GetSlotList;
+	list->C_GetSlotInfo = C_GetSlotInfo;
+	list->C_GetTokenInfo = C_GetTokenInfo;
+	list->C_GetMechanismList = C_GetMechanismList;
+	list->C_GetMechanismInfo = C_GetMechanismInfo;
+	list->C_InitToken = C_InitToken;
+	list->C_InitPIN = C_InitPIN;
+	list->C_SetPIN = C_SetPIN;
+	list->C_OpenSession = C_OpenSession;
+	list->C_CloseSession = C_CloseSession;
+	list->C_CloseAllSessions = C_CloseAllSessions;
+	list->C_GetSessionInfo = C_GetSessionInfo;
+	list->C_GetOperationState = C_GetOperationState;
+	list->C_SetOperationState = C_SetOperationState;
+	list->C_Login = C_Login;
+	list->C_Logout = C_Logout;
+	list->C_CreateObject = C_CreateObject;
+	list->C_CopyObject = C_CopyObject;
+	list->C_DestroyObject = C_DestroyObject;
+	list->C_GetObjectSize = C_GetObjectSize;
+	list->C_GetAttributeValue = C_GetAttributeValue;
+	list->C_SetAttributeValue = C_SetAttributeValue;
+	list->C_FindObjectsInit = C_FindObjectsInit;
+	list->C_FindObjects = C_FindObjects;
+	list->C_FindObjectsFinal = C_FindObjectsFinal;
+	list->C_EncryptInit = C_EncryptInit;
+	list->C_Encrypt = C_Encrypt;
+	list->C_EncryptUpdate = C_EncryptUpdate;
+	list->C_EncryptFinal = C_EncryptFinal;
+	list->C_DecryptInit = C_DecryptInit;
+	list->C_Decrypt = C_Decrypt;
+	list->C_DecryptUpdate = C_DecryptUpdate;
+	list->C_DecryptFinal = C_DecryptFinal;
+	list->C_DigestInit = C_DigestInit;
+	list->C_Digest = C_Digest;
+	list->C_DigestUpdate = C_DigestUpdate;
+	list->C_DigestKey = C_DigestKey;
+	list->C_DigestFinal = C_DigestFinal;
+	list->C_SignInit = C_SignInit;
+	list->C_Sign = C_Sign;
+	list->C_SignUpdate = C_SignUpdate;
+	list->C_SignFinal = C_SignFinal;
+	list->C_SignRecoverInit = C_SignRecoverInit;
+	list->C_SignRecover = C_SignRecover;
+	list->C_VerifyInit = C_VerifyInit;
+	list->C_Verify = C_Verify;
+	list->C_VerifyUpdate = C_VerifyUpdate;
+	list->C_VerifyFinal = C_VerifyFinal;
+	list->C_VerifyRecoverInit = C_VerifyRecoverInit;
+	list->C_VerifyRecover = C_VerifyRecover;
+	list->C_DigestEncryptUpdate = C_DigestEncryptUpdate;
+	list->C_DecryptDigestUpdate = C_DecryptDigestUpdate;
+	list->C_SignEncryptUpdate = C_SignEncryptUpdate;
+	list->C_DecryptVerifyUpdate = C_DecryptVerifyUpdate;
+	list->C_GenerateKey = C_GenerateKey;
+	list->C_GenerateKeyPair = C_GenerateKeyPair;
+	list->C_WrapKey = C_WrapKey;
+	list->C_UnwrapKey = C_UnwrapKey;
+	list->C_DeriveKey = C_DeriveKey;
+	list->C_SeedRandom = C_SeedRandom;
+	list->C_GenerateRandom = C_GenerateRandom;
+	list->C_GetFunctionStatus = C_GetFunctionStatus;
+	list->C_CancelFunction = C_CancelFunction;
+	list->C_WaitForSlotEvent = C_WaitForSlotEvent;
+	if (!v3) {
+		return list;
+	}
+
+	/* Add also PKCS #11 3.0 functions if requested and fixup version */
+	list_3_0 = (CK_FUNCTION_LIST_3_0_PTR) list;
+	list_3_0->version.major = 3;
+	list_3_0->version.minor = 0;
+	list_3_0->C_GetInterfaceList = C_GetInterfaceList;
+	list_3_0->C_GetInterface = C_GetInterface;
+	list_3_0->C_LoginUser = C_LoginUser;
+	list_3_0->C_SessionCancel = C_SessionCancel;
+	list_3_0->C_MessageEncryptInit = C_MessageEncryptInit;
+	list_3_0->C_EncryptMessage = C_EncryptMessage;
+	list_3_0->C_EncryptMessageBegin = C_EncryptMessageBegin;
+	list_3_0->C_EncryptMessageNext = C_EncryptMessageNext;
+	list_3_0->C_MessageEncryptFinal = C_MessageEncryptFinal;
+	list_3_0->C_MessageDecryptInit = C_MessageDecryptInit;
+	list_3_0->C_DecryptMessage = C_DecryptMessage;
+	list_3_0->C_DecryptMessageBegin = C_DecryptMessageBegin;
+	list_3_0->C_DecryptMessageNext = C_DecryptMessageNext;
+	list_3_0->C_MessageDecryptFinal = C_MessageDecryptFinal;
+	list_3_0->C_MessageSignInit = C_MessageSignInit;
+	list_3_0->C_SignMessage = C_SignMessage;
+	list_3_0->C_SignMessageBegin = C_SignMessageBegin;
+	list_3_0->C_SignMessageNext = C_SignMessageNext;
+	list_3_0->C_MessageSignFinal = C_MessageSignFinal;
+	list_3_0->C_MessageVerifyInit = C_MessageVerifyInit;
+	list_3_0->C_VerifyMessage = C_VerifyMessage;
+	list_3_0->C_VerifyMessageBegin = C_VerifyMessageBegin;
+	list_3_0->C_VerifyMessageNext = C_VerifyMessageNext;
+	list_3_0->C_MessageVerifyFinal = C_MessageVerifyFinal;
+
+	return list_3_0;
+}
+
+/* The compatibility interfaces that can be returned from Interface functions
+ * if the V3 API is used, but the proxied module does not support V3 API */
+#define NUM_INTERFACES 1
+CK_INTERFACE compat_interfaces[NUM_INTERFACES] = {
+	{"PKCS 11", NULL, 0}
+};
+
+/* Inits the spy. If successful, po != NULL */
 static CK_RV
 init_spy(void)
 {
+	CK_FUNCTION_LIST_PTR po_v2 = NULL;
 	const char *output, *module;
-	int rv = CKR_OK;
+	CK_RV rv = CKR_OK;
 #ifdef _WIN32
         char temp_path[PATH_MAX], expanded_path[PATH_MAX];
         DWORD temp_len, expanded_len;
@@ -64,83 +195,17 @@ init_spy(void)
 #endif
 
 	/* Allocates and initializes the pkcs11_spy structure */
-	pkcs11_spy = malloc(sizeof(CK_FUNCTION_LIST));
-	if (pkcs11_spy) {
-		/* with our own pkcs11.h we need to maintain this ourself */
-		pkcs11_spy->version.major = 2;
-		pkcs11_spy->version.major = 11;
-		pkcs11_spy->C_Initialize = C_Initialize;
-		pkcs11_spy->C_Finalize = C_Finalize;
-		pkcs11_spy->C_GetInfo = C_GetInfo;
-		pkcs11_spy->C_GetFunctionList = C_GetFunctionList;
-		pkcs11_spy->C_GetSlotList = C_GetSlotList;
-		pkcs11_spy->C_GetSlotInfo = C_GetSlotInfo;
-		pkcs11_spy->C_GetTokenInfo = C_GetTokenInfo;
-		pkcs11_spy->C_GetMechanismList = C_GetMechanismList;
-		pkcs11_spy->C_GetMechanismInfo = C_GetMechanismInfo;
-		pkcs11_spy->C_InitToken = C_InitToken;
-		pkcs11_spy->C_InitPIN = C_InitPIN;
-		pkcs11_spy->C_SetPIN = C_SetPIN;
-		pkcs11_spy->C_OpenSession = C_OpenSession;
-		pkcs11_spy->C_CloseSession = C_CloseSession;
-		pkcs11_spy->C_CloseAllSessions = C_CloseAllSessions;
-		pkcs11_spy->C_GetSessionInfo = C_GetSessionInfo;
-		pkcs11_spy->C_GetOperationState = C_GetOperationState;
-		pkcs11_spy->C_SetOperationState = C_SetOperationState;
-		pkcs11_spy->C_Login = C_Login;
-		pkcs11_spy->C_Logout = C_Logout;
-		pkcs11_spy->C_CreateObject = C_CreateObject;
-		pkcs11_spy->C_CopyObject = C_CopyObject;
-		pkcs11_spy->C_DestroyObject = C_DestroyObject;
-		pkcs11_spy->C_GetObjectSize = C_GetObjectSize;
-		pkcs11_spy->C_GetAttributeValue = C_GetAttributeValue;
-		pkcs11_spy->C_SetAttributeValue = C_SetAttributeValue;
-		pkcs11_spy->C_FindObjectsInit = C_FindObjectsInit;
-		pkcs11_spy->C_FindObjects = C_FindObjects;
-		pkcs11_spy->C_FindObjectsFinal = C_FindObjectsFinal;
-		pkcs11_spy->C_EncryptInit = C_EncryptInit;
-		pkcs11_spy->C_Encrypt = C_Encrypt;
-		pkcs11_spy->C_EncryptUpdate = C_EncryptUpdate;
-		pkcs11_spy->C_EncryptFinal = C_EncryptFinal;
-		pkcs11_spy->C_DecryptInit = C_DecryptInit;
-		pkcs11_spy->C_Decrypt = C_Decrypt;
-		pkcs11_spy->C_DecryptUpdate = C_DecryptUpdate;
-		pkcs11_spy->C_DecryptFinal = C_DecryptFinal;
-		pkcs11_spy->C_DigestInit = C_DigestInit;
-		pkcs11_spy->C_Digest = C_Digest;
-		pkcs11_spy->C_DigestUpdate = C_DigestUpdate;
-		pkcs11_spy->C_DigestKey = C_DigestKey;
-		pkcs11_spy->C_DigestFinal = C_DigestFinal;
-		pkcs11_spy->C_SignInit = C_SignInit;
-		pkcs11_spy->C_Sign = C_Sign;
-		pkcs11_spy->C_SignUpdate = C_SignUpdate;
-		pkcs11_spy->C_SignFinal = C_SignFinal;
-		pkcs11_spy->C_SignRecoverInit = C_SignRecoverInit;
-		pkcs11_spy->C_SignRecover = C_SignRecover;
-		pkcs11_spy->C_VerifyInit = C_VerifyInit;
-		pkcs11_spy->C_Verify = C_Verify;
-		pkcs11_spy->C_VerifyUpdate = C_VerifyUpdate;
-		pkcs11_spy->C_VerifyFinal = C_VerifyFinal;
-		pkcs11_spy->C_VerifyRecoverInit = C_VerifyRecoverInit;
-		pkcs11_spy->C_VerifyRecover = C_VerifyRecover;
-		pkcs11_spy->C_DigestEncryptUpdate = C_DigestEncryptUpdate;
-		pkcs11_spy->C_DecryptDigestUpdate = C_DecryptDigestUpdate;
-		pkcs11_spy->C_SignEncryptUpdate = C_SignEncryptUpdate;
-		pkcs11_spy->C_DecryptVerifyUpdate = C_DecryptVerifyUpdate;
-		pkcs11_spy->C_GenerateKey = C_GenerateKey;
-		pkcs11_spy->C_GenerateKeyPair = C_GenerateKeyPair;
-		pkcs11_spy->C_WrapKey = C_WrapKey;
-		pkcs11_spy->C_UnwrapKey = C_UnwrapKey;
-		pkcs11_spy->C_DeriveKey = C_DeriveKey;
-		pkcs11_spy->C_SeedRandom = C_SeedRandom;
-		pkcs11_spy->C_GenerateRandom = C_GenerateRandom;
-		pkcs11_spy->C_GetFunctionStatus = C_GetFunctionStatus;
-		pkcs11_spy->C_CancelFunction = C_CancelFunction;
-		pkcs11_spy->C_WaitForSlotEvent = C_WaitForSlotEvent;
-	}
-	else {
+	pkcs11_spy = allocate_function_list(0);
+	if (pkcs11_spy == NULL) {
 		return CKR_HOST_MEMORY;
 	}
+	pkcs11_spy_3_0 = allocate_function_list(1);
+	if (pkcs11_spy_3_0 == NULL) {
+		free(pkcs11_spy);
+		return CKR_HOST_MEMORY;
+	}
+
+	compat_interfaces[0].pFunctionList = pkcs11_spy;
 
 	/*
 	 * Don't use getenv() as the last parameter for scconf_get_str(),
@@ -151,9 +216,12 @@ init_spy(void)
 	if (output)
 		spy_output = fopen(output, "a");
 
+
+
+
 #ifdef _WIN32
 	if (!spy_output) {
-		/* try for the machine version first, as we may be runing
+		/* try for the machine version first, as we may be running
 		 * without a user during login
 		 */
 		rc = RegOpenKeyEx( HKEY_LOCAL_MACHINE, "Software\\OpenSC Project\\PKCS11-Spy", 0, KEY_QUERY_VALUE, &hKey );
@@ -172,7 +240,7 @@ init_spy(void)
 				}
 			}
 
-	                if( (rc == ERROR_SUCCESS) && (temp_len < PATH_MAX) )
+			if( (rc == ERROR_SUCCESS) && (temp_len < PATH_MAX) )
 				output = temp_path;
 			RegCloseKey( hKey );
 		}
@@ -188,7 +256,7 @@ init_spy(void)
 	module = getenv("PKCS11SPY");
 #ifdef _WIN32
 	if (!module) {
-		/* try for the machine version first, as we may be runing
+		/* try for the machine version first, as we may be running
 		 * without a user during login
 		 */
 		rc = RegOpenKeyEx( HKEY_LOCAL_MACHINE, "Software\\OpenSC Project\\PKCS11-Spy",
@@ -220,8 +288,8 @@ init_spy(void)
 		free(pkcs11_spy);
 		return CKR_DEVICE_ERROR;
 	}
-
-	modhandle = C_LoadModule(module, &po);
+	modhandle = C_LoadModule(module, &po_v2);
+	po = (CK_FUNCTION_LIST_3_0_PTR) po_v2;
 	if (modhandle && po) {
 		fprintf(spy_output, "Loaded: \"%s\"\n", module);
 	}
@@ -256,7 +324,7 @@ enter(const char *function)
 	gettimeofday (&tv, NULL);
 	tm = localtime (&tv.tv_sec);
 	strftime (time_string, sizeof(time_string), "%F %H:%M:%S", tm);
-	fprintf(spy_output, "%s.%03ld\n", time_string, tv.tv_usec / 1000);
+	fprintf(spy_output, "%s.%03ld\n", time_string, (long)tv.tv_usec / 1000);
 #endif
 
 }
@@ -332,6 +400,121 @@ spy_attribute_list_out(const char *name, CK_ATTRIBUTE_PTR pTemplate,
 	print_attribute_list(spy_output, pTemplate, ulCount);
 }
 
+static void
+spy_dump_mechanism_in(const char *name, CK_MECHANISM_PTR pMechanism)
+{
+	char param_name[64];
+
+	if (!pMechanism) {
+		fprintf(spy_output, "[in] %s = NULL\n", name);
+		return;
+	}
+
+	fprintf(spy_output, "[in] %s->type = %s\n", name, lookup_enum(MEC_T, pMechanism->mechanism));
+	switch (pMechanism->mechanism) {
+	case CKM_AES_GCM:
+		if (pMechanism->pParameter != NULL) {
+			CK_GCM_PARAMS *param =
+				(CK_GCM_PARAMS *) pMechanism->pParameter;
+			snprintf(param_name, sizeof(param_name), "%s->pParameter->pIv[ulIvLen]", name);
+			spy_dump_string_in(param_name,
+				param->pIv, param->ulIvLen);
+			snprintf(param_name, sizeof(param_name), "%s->pParameter->ulIvBits", name);
+			spy_dump_ulong_in(param_name, param->ulIvBits);
+			snprintf(param_name, sizeof(param_name), "%s->pParameter->pAAD[ulAADLen]", name);
+			spy_dump_string_in(param_name,
+				param->pAAD, param->ulAADLen);
+			fprintf(spy_output, "[in] %s->pParameter->ulTagBits = %lu\n", name, param->ulTagBits);
+		} else {
+			fprintf(spy_output, "[in] %s->pParameter = NULL\n", name);
+			break;
+		}
+		break;
+	case CKM_ECDH1_DERIVE:
+	case CKM_ECDH1_COFACTOR_DERIVE:
+		if (pMechanism->pParameter != NULL) {
+			CK_ECDH1_DERIVE_PARAMS *param =
+				(CK_ECDH1_DERIVE_PARAMS *) pMechanism->pParameter;
+			fprintf(spy_output, "[in] %s->pParameter->kdf = %s\n", name,
+				lookup_enum(CKD_T, param->kdf));
+			fprintf(spy_output, "[in] %s->pParameter->pSharedData[ulSharedDataLen] = ", name);
+			print_generic(spy_output, 0, param->pSharedData,
+				param->ulSharedDataLen, NULL);
+			fprintf(spy_output, "[in] %s->pParameter->pPublicData[ulPublicDataLen] = ", name);
+			print_generic(spy_output, 0, param->pPublicData,
+				param->ulPublicDataLen, NULL);
+		} else {
+			fprintf(spy_output, "[in] %s->pParameter = NULL\n", name);
+			break;
+		}
+		break;
+	case CKM_ECMQV_DERIVE:
+		if (pMechanism->pParameter != NULL) {
+			CK_ECMQV_DERIVE_PARAMS *param =
+				(CK_ECMQV_DERIVE_PARAMS *) pMechanism->pParameter;
+			fprintf(spy_output, "[in] %s->pParameter->kdf = %s\n", name,
+				lookup_enum(CKD_T, param->kdf));
+			fprintf(spy_output, "%s->pParameter->pSharedData[ulSharedDataLen] = ", name);
+			print_generic(spy_output, 0, param->pSharedData,
+				param->ulSharedDataLen, NULL);
+			fprintf(spy_output, "%s->pParameter->pPublicData[ulPublicDataLen] = ", name);
+			print_generic(spy_output, 0, param->pPublicData,
+				param->ulPublicDataLen, NULL);
+			fprintf(spy_output, "%s->pParameter->ulPrivateDataLen = %lu", name,
+				param->ulPrivateDataLen);
+			fprintf(spy_output, "%s->pParameter->hPrivateData = %lu", name, param->hPrivateData);
+			fprintf(spy_output, "%s->pParameter->pPublicData2[ulPublicDataLen2] = ", name);
+			print_generic(spy_output, 0, param->pPublicData2,
+				param->ulPublicDataLen2, NULL);
+			fprintf(spy_output, "%s->pParameter->publicKey = %lu", name, param->publicKey);
+		} else {
+			fprintf(spy_output, "[in] %s->pParameter = NULL\n", name);
+			break;
+		}
+		break;
+	case CKM_RSA_PKCS_OAEP:
+		if (pMechanism->pParameter != NULL) {
+			CK_RSA_PKCS_OAEP_PARAMS *param =
+				(CK_RSA_PKCS_OAEP_PARAMS *) pMechanism->pParameter;
+			fprintf(spy_output, "[in] %s->pParameter->hashAlg = %s\n", name,
+				lookup_enum(MEC_T, param->hashAlg));
+			fprintf(spy_output, "[in] %s->pParameter->mgf = %s\n", name,
+				lookup_enum(MGF_T, param->mgf));
+			fprintf(spy_output, "[in] %s->pParameter->source = %lu\n", name, param->source);
+			snprintf(param_name, sizeof(param_name), "%s->pParameter->pSourceData[ulSourceDalaLen]", name);
+			spy_dump_string_in(param_name,
+				param->pSourceData, param->ulSourceDataLen);
+		} else {
+			fprintf(spy_output, "[in] %s->pParameter = NULL\n", name);
+			break;
+		}
+		break;
+	case CKM_RSA_PKCS_PSS:
+	case CKM_SHA1_RSA_PKCS_PSS:
+	case CKM_SHA256_RSA_PKCS_PSS:
+	case CKM_SHA384_RSA_PKCS_PSS:
+	case CKM_SHA512_RSA_PKCS_PSS:
+		if (pMechanism->pParameter != NULL) {
+			CK_RSA_PKCS_PSS_PARAMS *param =
+				(CK_RSA_PKCS_PSS_PARAMS *) pMechanism->pParameter;
+			fprintf(spy_output, "[in] %s->pParameter->hashAlg = %s\n", name,
+				lookup_enum(MEC_T, param->hashAlg));
+			fprintf(spy_output, "[in] %s->pParameter->mgf = %s\n", name,
+				lookup_enum(MGF_T, param->mgf));
+			fprintf(spy_output, "[in] %s->pParameter->sLen = %lu\n", name,
+				param->sLen);
+		} else {
+			fprintf(spy_output, "[in] %s->pParameter = NULL\n", name);
+			break;
+		}
+		break;
+	default:
+		snprintf(param_name, sizeof(param_name), "%s->pParameter[ulParameterLen]", name);
+		spy_dump_string_in(param_name, pMechanism->pParameter, pMechanism->ulParameterLen);
+		break;
+	}
+}
+
 static void
 print_ptr_in(const char *name, CK_VOID_PTR ptr)
 {
@@ -348,6 +531,8 @@ CK_RV C_GetFunctionList
 	}
 
 	enter("C_GetFunctionList");
+	if (ppFunctionList == NULL)
+		return retne(CKR_ARGUMENTS_BAD);
 	*ppFunctionList = pkcs11_spy;
 	return retne(CKR_OK);
 }
@@ -477,9 +662,9 @@ C_GetMechanismInfo(CK_SLOT_ID  slotID, CK_MECHANISM_TYPE type,
 	enter("C_GetMechanismInfo");
 	spy_dump_ulong_in("slotID", slotID);
 	if (name)
-		fprintf(spy_output, "%30s \n", name);
+		fprintf(spy_output, "[in] type = %30s\n", name);
 	else
-		fprintf(spy_output, " Unknown Mechanism (%08lx)  \n", type);
+		fprintf(spy_output, "[in] type = Unknown Mechanism (%08lx)\n", type);
 
 	rv = po->C_GetMechanismInfo(slotID, type, pInfo);
 	if(rv == CKR_OK) {
@@ -538,10 +723,13 @@ C_OpenSession(CK_SLOT_ID  slotID, CK_FLAGS  flags, CK_VOID_PTR  pApplication,
 	enter("C_OpenSession");
 	spy_dump_ulong_in("slotID", slotID);
 	spy_dump_ulong_in("flags", flags);
-	fprintf(spy_output, "pApplication=%p\n", pApplication);
-	fprintf(spy_output, "Notify=%p\n", (void *)Notify);
+	fprintf(spy_output, "[in] pApplication = %p\n", pApplication);
+	fprintf(spy_output, "[in] Notify = %p\n", (void *)Notify);
 	rv = po->C_OpenSession(slotID, flags, pApplication, Notify, phSession);
-	spy_dump_ulong_out("*phSession", *phSession);
+	if (phSession)
+		spy_dump_ulong_out("*phSession", *phSession);
+	else
+		fprintf(spy_output, "[out] phSession = %p\n", phSession);
 	return retne(rv);
 }
 
@@ -785,7 +973,7 @@ C_EncryptInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT
 
 	enter("C_EncryptInit");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	spy_dump_mechanism_in("pMechanism", pMechanism);
 	spy_dump_ulong_in("hKey", hKey);
 	rv = po->C_EncryptInit(hSession, pMechanism, hKey);
 	return retne(rv);
@@ -843,7 +1031,7 @@ C_DecryptInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT
 
 	enter("C_DecryptInit");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	spy_dump_mechanism_in("pMechanism", pMechanism);
 	spy_dump_ulong_in("hKey", hKey);
 	rv = po->C_DecryptInit(hSession, pMechanism, hKey);
 	return retne(rv);
@@ -902,7 +1090,7 @@ C_DigestInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism)
 
 	enter("C_DigestInit");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	spy_dump_mechanism_in("pMechanism", pMechanism);
 	rv = po->C_DigestInit(hSession, pMechanism);
 	return retne(rv);
 }
@@ -968,7 +1156,7 @@ C_SignInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT_HA
 
 	enter("C_SignInit");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	spy_dump_mechanism_in("pMechanism", pMechanism);
 	spy_dump_ulong_in("hKey", hKey);
 	rv = po->C_SignInit(hSession, pMechanism, hKey);
 	return retne(rv);
@@ -1023,8 +1211,7 @@ C_SignRecoverInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OB
 
 	enter("C_SignRecoverInit");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n",
-			lookup_enum(MEC_T, pMechanism->mechanism));
+	spy_dump_mechanism_in("pMechanism", pMechanism);
 	spy_dump_ulong_in("hKey", hKey);
 	rv = po->C_SignRecoverInit(hSession, pMechanism, hKey);
 	return retne(rv);
@@ -1052,7 +1239,7 @@ C_VerifyInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT_
 
 	enter("C_VerifyInit");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	spy_dump_mechanism_in("pMechanism", pMechanism);
 	spy_dump_ulong_in("hKey", hKey);
 	rv = po->C_VerifyInit(hSession, pMechanism, hKey);
 	return retne(rv);
@@ -1105,7 +1292,7 @@ C_VerifyRecoverInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism,
 
 	enter("C_VerifyRecoverInit");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	spy_dump_mechanism_in("pMechanism", pMechanism);
 	spy_dump_ulong_in("hKey", hKey);
 	rv = po->C_VerifyRecoverInit(hSession, pMechanism, hKey);
 	return retne(rv);
@@ -1198,7 +1385,7 @@ C_GenerateKey(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism,
 
 	enter("C_GenerateKey");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	spy_dump_mechanism_in("pMechanism", pMechanism);
 	spy_attribute_list_in("pTemplate", pTemplate, ulCount);
 	rv = po->C_GenerateKey(hSession, pMechanism, pTemplate, ulCount, phKey);
 	if (rv == CKR_OK)
@@ -1217,7 +1404,7 @@ C_GenerateKeyPair(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism,
 
 	enter("C_GenerateKeyPair");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	spy_dump_mechanism_in("pMechanism", pMechanism);
 	spy_attribute_list_in("pPublicKeyTemplate", pPublicKeyTemplate, ulPublicKeyAttributeCount);
 	spy_attribute_list_in("pPrivateKeyTemplate", pPrivateKeyTemplate, ulPrivateKeyAttributeCount);
 	rv = po->C_GenerateKeyPair(hSession, pMechanism,
@@ -1240,7 +1427,7 @@ C_WrapKey(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism,
 
 	enter("C_WrapKey");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	spy_dump_mechanism_in("pMechanism", pMechanism);
 	spy_dump_ulong_in("hWrappingKey", hWrappingKey);
 	spy_dump_ulong_in("hKey", hKey);
 	rv = po->C_WrapKey(hSession, pMechanism, hWrappingKey, hKey, pWrappedKey, pulWrappedKeyLen);
@@ -1260,7 +1447,7 @@ C_UnwrapKey(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism,
 
 	enter("C_UnwrapKey");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	spy_dump_mechanism_in("pMechanism", pMechanism);
 	spy_dump_ulong_in("hUnwrappingKey", hUnwrappingKey);
 	spy_dump_string_in("pWrappedKey[ulWrappedKeyLen]", pWrappedKey, ulWrappedKeyLen);
 	spy_attribute_list_in("pTemplate", pTemplate, ulAttributeCount);
@@ -1279,7 +1466,7 @@ C_DeriveKey(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT_H
 
 	enter("C_DeriveKey");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	spy_dump_mechanism_in("pMechanism", pMechanism);
 	spy_dump_ulong_in("hBaseKey", hBaseKey);
 	spy_attribute_list_in("pTemplate", pTemplate, ulAttributeCount);
 	rv = po->C_DeriveKey(hSession, pMechanism, hBaseKey, pTemplate, ulAttributeCount, phKey);
@@ -1343,6 +1530,456 @@ C_WaitForSlotEvent(CK_FLAGS flags, CK_SLOT_ID_PTR pSlot, CK_VOID_PTR pRserved)
 
 	enter("C_WaitForSlotEvent");
 	spy_dump_ulong_in("flags", flags);
+	if (pSlot != NULL) {
+		spy_dump_ulong_in("pSlot", *pSlot);
+	}
 	rv = po->C_WaitForSlotEvent(flags, pSlot, pRserved);
 	return retne(rv);
 }
+
+/* PKCS #11 3.0 functions */
+static void
+spy_interface_function_list(CK_INTERFACE_PTR pInterface)
+{
+	CK_VERSION *version;
+
+	/* Do not touch unknown interfaces. We can not do anything with these */
+	if (strcmp(pInterface->pInterfaceName, "PKCS 11") != 0) {
+		return;
+	}
+
+	version = (CK_VERSION *)pInterface->pFunctionList;
+	if (version->major == 2) {
+		pInterface->pFunctionList = pkcs11_spy;
+	} else if (version->major == 3 && version->minor == 0) {
+		pInterface->pFunctionList = pkcs11_spy_3_0;
+	}
+}
+
+CK_RV
+C_GetInterfaceList(CK_INTERFACE_PTR pInterfacesList, CK_ULONG_PTR pulCount)
+{
+	CK_RV rv;
+
+	if (po == NULL) {
+		CK_RV rv = init_spy();
+		if (rv != CKR_OK)
+			return rv;
+	}
+
+	enter("C_GetInterfaceList");
+	if (po->version.major < 3) {
+		fprintf(spy_output, "[compat]\n");
+
+		memcpy(pInterfacesList, compat_interfaces, NUM_INTERFACES * sizeof(CK_INTERFACE));
+		*pulCount = NUM_INTERFACES;
+
+		spy_dump_desc_out("pInterfacesList");
+		print_interfaces_list(spy_output, pInterfacesList, *pulCount);
+		spy_dump_ulong_out("*pulCount", *pulCount);
+		return retne(CKR_OK);
+	}
+	rv = po->C_GetInterfaceList(pInterfacesList, pulCount);
+	if (rv == CKR_OK) {
+		spy_dump_desc_out("pInterfacesList");
+		print_interfaces_list(spy_output, pInterfacesList, *pulCount);
+		spy_dump_ulong_out("*pulCount", *pulCount);
+
+		/* Now, replace function lists of known interfaces (PKCS 11, v 2.x and 3.0) */
+		if (pInterfacesList != NULL) {
+			unsigned long i;
+			for (i = 0; i < *pulCount; i++) {
+				spy_interface_function_list(&pInterfacesList[i]);
+			}
+		}
+	}
+	return retne(rv);
+}
+
+CK_RV
+C_GetInterface(CK_UTF8CHAR_PTR pInterfaceName, CK_VERSION_PTR pVersion,
+	CK_INTERFACE_PTR_PTR ppInterface, CK_FLAGS flags)
+{
+	CK_RV rv;
+
+	if (po == NULL) {
+		CK_RV rv = init_spy();
+		if (rv != CKR_OK)
+			return rv;
+	}
+
+	enter("C_GetInterface");
+	if (po->version.major < 3) {
+		fprintf(spy_output, "[compat]\n");
+	}
+	spy_dump_string_in("pInterfaceName", pInterfaceName, strlen((char *)pInterfaceName));
+	if (pVersion != NULL) {
+		fprintf(spy_output, "[in] pVersion = %d.%d\n", pVersion->major, pVersion->minor);
+	} else {
+		fprintf(spy_output, "[in] pVersion = NULL\n");
+	}
+	fprintf(spy_output, "[in] flags = %s\n",
+		(flags & CKF_INTERFACE_FORK_SAFE ? "CKF_INTERFACE_FORK_SAFE" : ""));
+	if (po->version.major >= 3) {
+		rv = po->C_GetInterface(pInterfaceName, pVersion, ppInterface, flags);
+		if (ppInterface != NULL) {
+			spy_interface_function_list(*ppInterface);
+		}
+	} else {
+		if ((pInterfaceName == NULL_PTR || strcmp((char *)pInterfaceName, "PKCS 11") == 0) &&
+			(pVersion == NULL_PTR || (pVersion->major == 2 && pVersion->minor == 11)) &&
+			flags == 0) {
+			*ppInterface = &compat_interfaces[0];
+			return retne(CKR_OK);
+		}
+		/* We can not serve this particular interface */
+		return retne(CKR_ARGUMENTS_BAD);
+	}
+
+	return retne(rv);
+}
+
+CK_RV
+C_LoginUser(CK_SESSION_HANDLE hSession, CK_USER_TYPE userType,
+	CK_CHAR_PTR pPin, CK_ULONG ulPinLen, CK_UTF8CHAR_PTR pUsername, CK_ULONG ulUsernameLen)
+{
+	CK_RV rv;
+
+	enter("C_LoginUser");
+	spy_dump_ulong_in("hSession", hSession);
+	fprintf(spy_output, "[in] userType = %s\n",
+			lookup_enum(USR_T, userType));
+	spy_dump_string_in("pPin[ulPinLen]", pPin, ulPinLen);
+	spy_dump_string_in("pUsername[ulUsernameLen]", pUsername, ulUsernameLen);
+	rv = po->C_LoginUser(hSession, userType, pPin, ulPinLen, pUsername, ulUsernameLen);
+
+	return retne(rv);
+}
+
+CK_RV C_SessionCancel(CK_SESSION_HANDLE hSession, CK_FLAGS flags)
+{
+	CK_RV rv;
+
+	enter("C_SessionCancel");
+	spy_dump_ulong_in("hSession", hSession);
+	fprintf(spy_output, "[in] flags = %s%s%s%s%s%s%s%s%s%s%s%s\n",
+		(flags & CKF_ENCRYPT)           ? "Encrypt "  : "",
+		(flags & CKF_DECRYPT)           ? "Decrypt "  : "",
+		(flags & CKF_DIGEST)            ? "Digest "   : "",
+		(flags & CKF_SIGN)              ? "Sign "     : "",
+		(flags & CKF_SIGN_RECOVER)      ? "SigRecov " : "",
+		(flags & CKF_VERIFY)            ? "Verify "   : "",
+		(flags & CKF_VERIFY_RECOVER)    ? "VerRecov " : "",
+		(flags & CKF_GENERATE)          ? "Generate " : "",
+		(flags & CKF_GENERATE_KEY_PAIR) ? "KeyPair "  : "",
+		(flags & CKF_WRAP)              ? "Wrap "     : "",
+		(flags & CKF_UNWRAP)            ? "Unwrap "   : "",
+		(flags & CKF_DERIVE)            ? "Derive "   : "");
+	rv = po->C_SessionCancel(hSession, flags);
+
+	return retne(rv);
+}
+
+CK_RV
+C_MessageEncryptInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT_HANDLE hKey)
+{
+	CK_RV rv;
+
+	enter("C_MessageEncryptInit");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_mechanism_in("pMechanism", pMechanism);
+	spy_dump_ulong_in("hKey", hKey);
+	rv = po->C_MessageEncryptInit(hSession, pMechanism, hKey);
+	return retne(rv);
+}
+
+CK_RV
+C_EncryptMessage(CK_SESSION_HANDLE hSession, CK_VOID_PTR pParameter, CK_ULONG ulParameterLen,
+	CK_BYTE_PTR pAssociatedData, CK_ULONG ulAssociatedDataLen,
+	CK_BYTE_PTR pPlaintext, CK_ULONG ulPlaintextLen,
+	CK_BYTE_PTR pCiphertext, CK_ULONG_PTR pulCiphertextLen)
+{
+	CK_RV rv;
+
+	enter("C_EncryptMessage");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_string_in("pParameter[ulParameterLen]", pParameter, ulParameterLen);
+	spy_dump_string_in("pAssociatedData[ulAssociatedDataLen]", pAssociatedData, ulAssociatedDataLen);
+	spy_dump_string_in("pPlaintext[ulPlaintextLen]", pPlaintext, ulPlaintextLen);
+	rv = po->C_EncryptMessage(hSession, pParameter, ulParameterLen,
+		pAssociatedData, ulAssociatedDataLen, pPlaintext, ulPlaintextLen,
+		pCiphertext, pulCiphertextLen);
+	if (rv == CKR_OK) {
+		spy_dump_string_out("pCiphertext[*pulCiphertextLen]", pCiphertext, *pulCiphertextLen);
+	}
+	return retne(rv);
+}
+
+CK_RV
+C_EncryptMessageBegin(CK_SESSION_HANDLE hSession, CK_VOID_PTR pParameter, CK_ULONG ulParameterLen,
+	CK_BYTE_PTR pAssociatedData, CK_ULONG ulAssociatedDataLen)
+{
+	CK_RV rv;
+
+	enter("C_EncryptMessageBegin");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_string_in("pParameter[ulParameterLen]", pParameter, ulParameterLen);
+	spy_dump_string_in("pAssociatedData[ulAssociatedDataLen]", pAssociatedData, ulAssociatedDataLen);
+	rv = po->C_EncryptMessageBegin(hSession, pParameter, ulParameterLen,
+		pAssociatedData, ulAssociatedDataLen);
+	return retne(rv);
+}
+
+CK_RV
+C_EncryptMessageNext(CK_SESSION_HANDLE hSession, CK_VOID_PTR pParameter, CK_ULONG ulParameterLen,
+	CK_BYTE_PTR pPlaintextPart, CK_ULONG ulPlaintextPartLen,
+	CK_BYTE_PTR pCiphertextPart, CK_ULONG_PTR pulCiphertextPartLen, CK_FLAGS flags)
+{
+	CK_RV rv;
+
+	enter("C_EncryptMessageNext");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_string_in("pParameter[ulParameterLen]", pParameter, ulParameterLen);
+	spy_dump_string_in("pPlaintextPart[ulPlaintextPartLen]", pPlaintextPart, ulPlaintextPartLen);
+	rv = po->C_EncryptMessageNext(hSession, pParameter, ulParameterLen,
+		pPlaintextPart, ulPlaintextPartLen, pCiphertextPart, pulCiphertextPartLen, flags);
+	if (rv == CKR_OK) {
+		spy_dump_string_out("pCiphertextPart[*pulCiphertextPartLen]",
+			pCiphertextPart, *pulCiphertextPartLen);
+	}
+	fprintf(spy_output, "[in] flags = %s\n",
+		(flags & CKF_END_OF_MESSAGE ? "CKF_END_OF_MESSAGE" : ""));
+	return retne(rv);
+}
+
+CK_RV
+C_MessageEncryptFinal(CK_SESSION_HANDLE hSession)
+{
+	CK_RV rv;
+
+	enter("C_MessageEncryptFinal");
+	spy_dump_ulong_in("hSession", hSession);
+	rv = po->C_MessageEncryptFinal(hSession);
+	return retne(rv);
+}
+
+CK_RV
+C_MessageDecryptInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT_HANDLE hKey)
+{
+	CK_RV rv;
+
+	enter("C_MessageDecryptInit");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_mechanism_in("pMechanism", pMechanism);
+	spy_dump_ulong_in("hKey", hKey);
+	rv = po->C_MessageDecryptInit(hSession, pMechanism, hKey);
+	return retne(rv);
+}
+
+CK_RV
+C_DecryptMessage(CK_SESSION_HANDLE hSession, CK_VOID_PTR pParameter, CK_ULONG ulParameterLen,
+	CK_BYTE_PTR pAssociatedData, CK_ULONG ulAssociatedDataLen,
+	CK_BYTE_PTR pCiphertext, CK_ULONG ulCiphertextLen,
+	CK_BYTE_PTR pPlaintext, CK_ULONG_PTR pulPlaintextLen)
+{
+	CK_RV rv;
+
+	enter("C_DecryptMessage");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_string_in("pParameter[ulParameterLen]", pParameter, ulParameterLen);
+	spy_dump_string_in("pAssociatedData[ulAssociatedDataLen]", pAssociatedData, ulAssociatedDataLen);
+	spy_dump_string_in("pCiphertext[ulCiphertextLen]", pCiphertext, ulCiphertextLen);
+	rv = po->C_DecryptMessage(hSession, pParameter, ulParameterLen,
+		pAssociatedData, ulAssociatedDataLen, pCiphertext, ulCiphertextLen,
+		pPlaintext, pulPlaintextLen);
+	if (rv == CKR_OK) {
+		spy_dump_string_out("pPlaintext[*pulPlaintextLen]", pPlaintext, *pulPlaintextLen);
+	}
+	return retne(rv);
+}
+
+CK_RV
+C_DecryptMessageBegin(CK_SESSION_HANDLE hSession, CK_VOID_PTR pParameter, CK_ULONG ulParameterLen,
+	CK_BYTE_PTR pAssociatedData, CK_ULONG ulAssociatedDataLen)
+{
+	CK_RV rv;
+
+	enter("C_DecryptMessageBegin");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_string_in("pParameter[ulParameterLen]", pParameter, ulParameterLen);
+	spy_dump_string_in("pAssociatedData[ulAssociatedDataLen]", pAssociatedData, ulAssociatedDataLen);
+	rv = po->C_DecryptMessageBegin(hSession, pParameter, ulParameterLen,
+		pAssociatedData, ulAssociatedDataLen);
+	return retne(rv);
+}
+
+CK_RV
+C_DecryptMessageNext(CK_SESSION_HANDLE hSession, CK_VOID_PTR pParameter, CK_ULONG ulParameterLen,
+	CK_BYTE_PTR pCiphertextPart, CK_ULONG ulCiphertextPartLen,
+	CK_BYTE_PTR pPlaintextPart, CK_ULONG_PTR pulPlaintextPartLen, CK_FLAGS flags)
+{
+	CK_RV rv;
+
+	enter("C_DecryptMessageNext");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_string_in("pParameter[ulParameterLen]", pParameter, ulParameterLen);
+	spy_dump_string_in("pCiphertextPart[ulCiphertextPartLen]", pCiphertextPart, ulCiphertextPartLen);
+	rv = po->C_DecryptMessageNext(hSession, pParameter, ulParameterLen,
+		pCiphertextPart, ulCiphertextPartLen, pPlaintextPart, pulPlaintextPartLen, flags);
+	if (rv == CKR_OK) {
+		spy_dump_string_out("pPlaintextPart[*pulPlaintextPartLen]",
+			pPlaintextPart, *pulPlaintextPartLen);
+	}
+	fprintf(spy_output, "[in] flags = %s\n",
+		(flags & CKF_END_OF_MESSAGE ? "CKF_END_OF_MESSAGE" : ""));
+	return retne(rv);
+}
+
+CK_RV
+C_MessageDecryptFinal(CK_SESSION_HANDLE hSession)
+{
+	CK_RV rv;
+
+	enter("C_MessageDecryptFinal");
+	spy_dump_ulong_in("hSession", hSession);
+	rv = po->C_MessageDecryptFinal(hSession);
+	return retne(rv);
+}
+
+CK_RV
+C_MessageSignInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT_HANDLE hKey)
+{
+	CK_RV rv;
+
+	enter("C_MessageSignInit");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_mechanism_in("pMechanism", pMechanism);
+	spy_dump_ulong_in("hKey", hKey);
+	rv = po->C_MessageSignInit(hSession, pMechanism, hKey);
+	return retne(rv);
+}
+
+CK_RV
+C_SignMessage(CK_SESSION_HANDLE hSession, CK_VOID_PTR pParameter, CK_ULONG ulParameterLen,
+	CK_BYTE_PTR pData, CK_ULONG ulDataLen, CK_BYTE_PTR pSignature, CK_ULONG_PTR pulSignatureLen)
+{
+	CK_RV rv;
+
+	enter("C_SignMessage");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_string_in("pParameter[ulParameterLen]", pParameter, ulParameterLen);
+	spy_dump_string_in("pData[ulDataLen]", pData, ulDataLen);
+	rv = po->C_SignMessage(hSession, pParameter, ulParameterLen,
+		pData, ulDataLen, pSignature, pulSignatureLen);
+	if (rv == CKR_OK) {
+		spy_dump_string_out("pSignature[*pulSignatureLen]", pSignature, *pulSignatureLen);
+	}
+	return retne(rv);
+}
+
+CK_RV
+C_SignMessageBegin(CK_SESSION_HANDLE hSession, CK_VOID_PTR pParameter, CK_ULONG ulParameterLen)
+{
+	CK_RV rv;
+
+	enter("C_SignMessageBegin");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_string_in("pParameter[ulParameterLen]", pParameter, ulParameterLen);
+	rv = po->C_SignMessageBegin(hSession, pParameter, ulParameterLen);
+	return retne(rv);
+}
+
+CK_RV
+C_SignMessageNext(CK_SESSION_HANDLE hSession, CK_VOID_PTR pParameter, CK_ULONG ulParameterLen,
+	CK_BYTE_PTR pData, CK_ULONG ulDataLen, CK_BYTE_PTR pSignature, CK_ULONG_PTR pulSignatureLen)
+{
+	CK_RV rv;
+
+	enter("C_SignMessageNext");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_string_in("pParameter[ulParameterLen]", pParameter, ulParameterLen);
+	spy_dump_string_in("pData[ulDataLen]", pData, ulDataLen);
+	rv = po->C_SignMessageNext(hSession, pParameter, ulParameterLen,
+		pData, ulDataLen, pSignature, pulSignatureLen);
+	if (rv == CKR_OK) {
+		spy_dump_string_out("pSignature[*pulSignatureLen]", pSignature, *pulSignatureLen);
+	}
+	return retne(rv);
+}
+
+CK_RV
+C_MessageSignFinal(CK_SESSION_HANDLE hSession)
+{
+	CK_RV rv;
+
+	enter("C_MessageSignFinal");
+	spy_dump_ulong_in("hSession", hSession);
+	rv = po->C_MessageSignFinal(hSession);
+	return retne(rv);
+}
+
+CK_RV
+C_MessageVerifyInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT_HANDLE  hKey)
+{
+	CK_RV rv;
+
+	enter("C_MessageVerifyInit");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_mechanism_in("pMechanism", pMechanism);
+	spy_dump_ulong_in("hKey", hKey);
+	rv = po->C_MessageVerifyInit(hSession, pMechanism, hKey);
+	return retne(rv);
+}
+
+CK_RV
+C_VerifyMessage(CK_SESSION_HANDLE hSession, CK_VOID_PTR pParameter, CK_ULONG ulParameterLen,
+	CK_BYTE_PTR pData, CK_ULONG ulDataLen, CK_BYTE_PTR pSignature, CK_ULONG ulSignatureLen)
+{
+	CK_RV rv;
+
+	enter("C_VerifyMessage");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_string_in("pParameter[ulParameterLen]", pParameter, ulParameterLen);
+	spy_dump_string_in("pData[ulDataLen]", pData, ulDataLen);
+	spy_dump_string_in("pSignature[ulSignatureLen]", pSignature, ulSignatureLen);
+	rv = po->C_VerifyMessage(hSession, pParameter, ulParameterLen,
+		pData, ulDataLen, pSignature, ulSignatureLen);
+	return retne(rv);
+}
+
+CK_RV
+C_VerifyMessageBegin(CK_SESSION_HANDLE hSession, CK_VOID_PTR pParameter, CK_ULONG ulParameterLen)
+{
+	CK_RV rv;
+
+	enter("C_VerifyMessageBegin");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_string_in("pParameter[ulParameterLen]", pParameter, ulParameterLen);
+	rv = po->C_VerifyMessageBegin(hSession, pParameter, ulParameterLen);
+	return retne(rv);
+}
+
+CK_RV C_VerifyMessageNext(CK_SESSION_HANDLE hSession, CK_VOID_PTR pParameter, CK_ULONG ulParameterLen,
+	CK_BYTE_PTR pData, CK_ULONG ulDataLen, CK_BYTE_PTR pSignature, CK_ULONG ulSignatureLen)
+{
+	CK_RV rv;
+
+	enter("C_VerifyMessageNext");
+	spy_dump_ulong_in("hSession", hSession);
+	spy_dump_string_in("pParameter[ulParameterLen]", pParameter, ulParameterLen);
+	spy_dump_string_in("pData[ulDataLen]", pData, ulDataLen);
+	spy_dump_string_in("pSignature[ulSignatureLen]", pSignature, ulSignatureLen);
+	rv = po->C_VerifyMessageNext(hSession, pParameter, ulParameterLen,
+		pData, ulDataLen, pSignature, ulSignatureLen);
+	return retne(rv);
+}
+
+CK_RV C_MessageVerifyFinal(CK_SESSION_HANDLE hSession)
+{
+	CK_RV rv;
+
+	enter("C_MessageVerifyFinal");
+	spy_dump_ulong_in("hSession", hSession);
+	rv = po->C_MessageVerifyFinal(hSession);
+	return retne(rv);
+}
