@@ -27,6 +27,32 @@
 
 #include "sc-pkcs11.h"
 
+/* Print virtual_slots list. Called by DEBUG_VSS(S, C) */
+void _debug_virtual_slots(sc_pkcs11_slot_t *p)
+{
+	int i, vs_size;
+	sc_pkcs11_slot_t * slot;
+
+	vs_size = list_size(&virtual_slots);
+	_sc_debug(context, 10,
+			"VSS size:%d", vs_size);
+	_sc_debug(context, 10,
+			"VSS  [i] id   flags LU events nsessions slot_info.flags reader p11card description");
+	for (i = 0; i < vs_size; i++) {
+		slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, i);
+		if (slot) {
+			_sc_debug(context, 10,
+				"VSS %s[%d] 0x%2.2lx 0x%4.4x %d  %d  %d %4.4lx  %p %p %.64s",
+				((slot == p) ? "*" : " "),
+				i, slot->id, slot->flags, slot->login_user, slot->events, slot->nsessions,
+				slot->slot_info.flags,
+				slot->reader, slot->p11card,
+				slot->slot_info.slotDescription);
+		}
+	}
+	_sc_debug(context, 10, "VSS END");
+}
+
 static struct sc_pkcs11_framework_ops *frameworks[] = {
 	&framework_pkcs15,
 #ifdef USE_PKCS15_INIT
@@ -37,26 +63,43 @@ static struct sc_pkcs11_framework_ops *frameworks[] = {
 	NULL
 };
 
-static struct sc_pkcs11_slot * reader_get_slot(sc_reader_t *reader)
+static struct sc_pkcs11_slot * reader_reclaim_slot(sc_reader_t *reader)
 {
 	unsigned int i;
+	CK_UTF8CHAR slotDescription[64];
+	CK_UTF8CHAR manufacturerID[32];
+
+	strcpy_bp(slotDescription, reader->name, 64);
+	strcpy_bp(manufacturerID, reader->vendor, 32);
 
 	/* Locate a slot related to the reader */
 	for (i = 0; i<list_size(&virtual_slots); i++) {
 		sc_pkcs11_slot_t *slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, i);
-		if (slot->reader == reader)
+		if (slot->reader == NULL && reader != NULL
+				&& 0 == memcmp(slot->slot_info.slotDescription, slotDescription, 64)
+				&& 0 == memcmp(slot->slot_info.manufacturerID, manufacturerID, 32)
+				&& slot->slot_info.hardwareVersion.major == reader->version_major
+				&& slot->slot_info.hardwareVersion.minor == reader->version_minor) {
 			return slot;
+		}
 	}
 	return NULL;
 }
 
-static void init_slot_info(CK_SLOT_INFO_PTR pInfo)
+void init_slot_info(CK_SLOT_INFO_PTR pInfo, sc_reader_t *reader)
 {
-	strcpy_bp(pInfo->slotDescription, "Virtual hotplug slot", 64);
-	strcpy_bp(pInfo->manufacturerID, OPENSC_VS_FF_COMPANY_NAME, 32);
+	if (reader) {
+		strcpy_bp(pInfo->slotDescription, reader->name, 64);
+		strcpy_bp(pInfo->manufacturerID, reader->vendor, 32);
+		pInfo->hardwareVersion.major = reader->version_major;
+		pInfo->hardwareVersion.minor = reader->version_minor;
+	} else {
+		strcpy_bp(pInfo->slotDescription, "Virtual hotplug slot", 64);
+		strcpy_bp(pInfo->manufacturerID, OPENSC_VS_FF_COMPANY_NAME, 32);
+		pInfo->hardwareVersion.major = OPENSC_VERSION_MAJOR;
+		pInfo->hardwareVersion.minor = OPENSC_VERSION_MINOR;
+	}
 	pInfo->flags = CKF_REMOVABLE_DEVICE | CKF_HW_SLOT;
-	pInfo->hardwareVersion.major = OPENSC_VERSION_MAJOR;
-	pInfo->hardwareVersion.minor = OPENSC_VERSION_MINOR;
 	pInfo->firmwareVersion.major = 0;
 	pInfo->firmwareVersion.minor = 0;
 }
@@ -75,86 +118,69 @@ static int object_list_seeker(const void *el, const void *key)
 
 CK_RV create_slot(sc_reader_t *reader)
 {
-	struct sc_pkcs11_slot *slot;
+	/* find unused slots previously allocated for the same reader */
+	struct sc_pkcs11_slot *slot = reader_reclaim_slot(reader);
 
-	if (list_size(&virtual_slots) >= sc_pkcs11_conf.max_virtual_slots)
-		return CKR_FUNCTION_FAILED;
+	/* create a new slot if no empty slot is available */
+	if (!slot) {
+		sc_log(context, "Creating new slot");
+		if (list_size(&virtual_slots) >= sc_pkcs11_conf.max_virtual_slots)
+			return CKR_FUNCTION_FAILED;
 
-	slot = (struct sc_pkcs11_slot *)calloc(1, sizeof(struct sc_pkcs11_slot));
-	if (!slot)
-		return CKR_HOST_MEMORY;
+		slot = (struct sc_pkcs11_slot *)calloc(1, sizeof(struct sc_pkcs11_slot));
+		if (!slot)
+			return CKR_HOST_MEMORY;
+
+		list_append(&virtual_slots, slot);
+		if (0 != list_init(&slot->objects)) {
+			return CKR_HOST_MEMORY;
+		}
+		list_attributes_seeker(&slot->objects, object_list_seeker);
+
+		if (0 != list_init(&slot->logins)) {
+			return CKR_HOST_MEMORY;
+		}
+	} else {
+		DEBUG_VSS(slot, "Reusing this old slot");
+
+		/* reuse the old list of logins/objects since they should be empty */
+		list_t logins = slot->logins;
+		list_t objects = slot->objects;
+
+		memset(slot, 0, sizeof *slot);
+
+		slot->logins = logins;
+		slot->objects = objects;
+	}
 
-	list_append(&virtual_slots, slot);
 	slot->login_user = -1;
 	slot->id = (CK_SLOT_ID) list_locate(&virtual_slots, slot);
-	sc_log(context, "Creating slot with id 0x%lx", slot->id);
+	init_slot_info(&slot->slot_info, reader);
+	slot->reader = reader;
 
-	list_init(&slot->objects);
-	list_attributes_seeker(&slot->objects, object_list_seeker);
-
-	list_init(&slot->logins);
-
-	init_slot_info(&slot->slot_info);
-	if (reader != NULL) {
-		slot->reader = reader;
-		strcpy_bp(slot->slot_info.manufacturerID, reader->vendor, 32);
-		strcpy_bp(slot->slot_info.slotDescription, reader->name, 64);
-		slot->slot_info.hardwareVersion.major = reader->version_major;
-		slot->slot_info.hardwareVersion.minor = reader->version_minor;
-	}
+	DEBUG_VSS(slot, "Finished initializing this slot");
 
 	return CKR_OK;
 }
 
-void delete_slot(struct sc_pkcs11_slot *slot)
+void sc_pkcs11_card_free(struct sc_pkcs11_card *p11card)
 {
-	if (slot) {
-		list_destroy(&slot->objects);
-		list_destroy(&slot->logins);
-		list_delete(&virtual_slots, slot);
-		free(slot);
-	}
-}
-
-
-/* create slots associated with a reader, called whenever a reader is seen. */
-CK_RV initialize_reader(sc_reader_t *reader)
-{
-	unsigned int i;
-	CK_RV rv;
-
-	scconf_block *conf_block = NULL;
-	const scconf_list *list = NULL;
-
-	conf_block = sc_get_conf_block(context, "pkcs11", NULL, 1);
-	if (conf_block != NULL) {
-		list = scconf_find_list(conf_block, "ignored_readers");
-		while (list != NULL) {
-			if (strstr(reader->name, list->data) != NULL) {
-				sc_log(context, "Ignoring reader \'%s\' because of \'%s\'\n", reader->name, list->data);
-				return CKR_OK;
+	if (p11card) {
+		size_t i;
+		if (p11card->framework && p11card->framework->unbind)
+			p11card->framework->unbind(p11card);
+		sc_disconnect_card(p11card->card);
+		for (i=0; i < p11card->nmechanisms; ++i) {
+			if (p11card->mechanisms[i]->free_mech_data) {
+				p11card->mechanisms[i]->free_mech_data(p11card->mechanisms[i]->mech_data);
 			}
-			list = list->next;
+			free(p11card->mechanisms[i]);
 		}
+		free(p11card->mechanisms);
+		free(p11card);
 	}
-
-	for (i = 0; i < sc_pkcs11_conf.slots_per_card; i++) {
-		rv = create_slot(reader);
-		if (rv != CKR_OK)
-			return rv;
-	}
-
-	sc_log(context, "Initialize reader '%s': detect SC card presence", reader->name);
-	if (sc_detect_card_presence(reader))   {
-		sc_log(context, "Initialize reader '%s': detect PKCS11 card presence", reader->name);
-		card_detect(reader);
-	}
-
-	sc_log(context, "Reader '%s' initialized", reader->name);
-	return CKR_OK;
 }
 
-
 CK_RV card_removed(sc_reader_t * reader)
 {
 	unsigned int i;
@@ -173,18 +199,7 @@ CK_RV card_removed(sc_reader_t * reader)
 		}
 	}
 
-	if (p11card) {
-		p11card->framework->unbind(p11card);
-		sc_disconnect_card(p11card->card);
-		for (i=0; i < p11card->nmechanisms; ++i) {
-			if (p11card->mechanisms[i]->free_mech_data) {
-				p11card->mechanisms[i]->free_mech_data(p11card->mechanisms[i]->mech_data);
-			}
-			free(p11card->mechanisms[i]);
-		}
-		free(p11card->mechanisms);
-		free(p11card);
-	}
+	sc_pkcs11_card_free(p11card);
 
 	return CKR_OK;
 }
@@ -193,13 +208,12 @@ CK_RV card_removed(sc_reader_t * reader)
 CK_RV card_detect(sc_reader_t *reader)
 {
 	struct sc_pkcs11_card *p11card = NULL;
+	int free_p11card = 0;
 	int rc;
 	CK_RV rv;
 	unsigned int i;
 	int j;
 
-	rv = CKR_OK;
-
 	sc_log(context, "%s: Detecting smart card", reader->name);
 	/* Check if someone inserted a card */
 again:
@@ -241,15 +255,29 @@ again:
 		p11card = (struct sc_pkcs11_card *)calloc(1, sizeof(struct sc_pkcs11_card));
 		if (!p11card)
 			return CKR_HOST_MEMORY;
+		free_p11card = 1;
 		p11card->reader = reader;
 	}
 
 	if (p11card->card == NULL) {
 		sc_log(context, "%s: Connecting ... ", reader->name);
 		rc = sc_connect_card(reader, &p11card->card);
-		if (rc != SC_SUCCESS)   {
+		if (rc != SC_SUCCESS) {
 			sc_log(context, "%s: SC connect card error %i", reader->name, rc);
-			return sc_to_cryptoki_error(rc, NULL);
+			rv = sc_to_cryptoki_error(rc, NULL);
+			goto fail;
+		}
+
+		/* escape commands are only guaranteed to be working with a card
+		 * inserted. That's why by now, after sc_connect_card() the reader's
+		 * metadata may have changed. We re-initialize the metadata for every
+		 * slot of this reader here. */
+		if (reader->flags & SC_READER_ENABLE_ESCAPE) {
+			for (i = 0; i<list_size(&virtual_slots); i++) {
+				sc_pkcs11_slot_t *slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, i);
+				if (slot->reader == reader)
+					init_slot_info(&slot->slot_info, reader);
+			}
 		}
 
 		sc_log(context, "%s: Connected SC card %p", reader->name, p11card->card);
@@ -266,8 +294,10 @@ again:
 			if (frameworks[i]->bind != NULL)
 				break;
 		/*TODO: only first framework is used: pkcs15init framework is not reachable here */
-		if (frameworks[i] == NULL)
-			return CKR_GENERAL_ERROR;
+		if (frameworks[i] == NULL) {
+			rv = CKR_GENERAL_ERROR;
+			goto fail;
+		}
 
 		p11card->framework = frameworks[i];
 
@@ -275,12 +305,17 @@ again:
 		sc_log(context, "%s: Detected framework %d. Creating tokens.", reader->name, i);
 		/* Bind 'generic' application or (emulated?) card without applications */
 		if (app_generic || !p11card->card->app_count)   {
-			scconf_block *atrblock = NULL;
+			scconf_block *conf_block = NULL;
 			int enable_InitToken = 0;
 
-			atrblock = sc_match_atr_block(p11card->card->ctx, NULL, &p11card->reader->atr);
-			if (atrblock)
-				enable_InitToken = scconf_get_bool(atrblock, "pkcs11_enable_InitToken", 0);
+			conf_block = sc_match_atr_block(p11card->card->ctx, NULL,
+				&p11card->reader->atr);
+			if (!conf_block) /* check default block */
+				conf_block = sc_get_conf_block(context,
+					"framework", "pkcs15", 1);
+
+			enable_InitToken = scconf_get_bool(conf_block,
+				"pkcs11_enable_InitToken", 0);
 
 			sc_log(context, "%s: Try to bind 'generic' token.", reader->name);
 			rv = frameworks[i]->bind(p11card, app_generic);
@@ -289,16 +324,22 @@ again:
 				rv = CKR_OK;
 			}
 			if (rv != CKR_OK)   {
-				sc_log(context, "%s: cannot bind 'generic' token: rv 0x%X", reader->name, rv);
-				return rv;
+				sc_log(context,
+				       "%s: cannot bind 'generic' token: rv 0x%lX",
+				       reader->name, rv);
+				goto fail;
 			}
 
 			sc_log(context, "%s: Creating 'generic' token.", reader->name);
 			rv = frameworks[i]->create_tokens(p11card, app_generic);
 			if (rv != CKR_OK)   {
-				sc_log(context, "%s: create 'generic' token error 0x%X", reader->name, rv);
-				return rv;
+				sc_log(context,
+				       "%s: create 'generic' token error 0x%lX",
+				       reader->name, rv);
+				goto fail;
 			}
+			/* p11card is now bound to some slot */
+			free_p11card = 0;
 		}
 
 		/* Now bind the rest of applications that are not 'generic' */
@@ -312,46 +353,78 @@ again:
 			sc_log(context, "%s: Binding %s token.", reader->name, app_name);
 			rv = frameworks[i]->bind(p11card, app_info);
 			if (rv != CKR_OK)   {
-				sc_log(context, "%s: bind %s token error Ox%X", reader->name, app_name, rv);
+				sc_log(context, "%s: bind %s token error Ox%lX",
+				       reader->name, app_name, rv);
 				continue;
 			}
 
 			sc_log(context, "%s: Creating %s token.", reader->name, app_name);
 			rv = frameworks[i]->create_tokens(p11card, app_info);
 			if (rv != CKR_OK)   {
-				sc_log(context, "%s: create %s token error 0x%X", reader->name, app_name, rv);
-				return rv;
+				sc_log(context,
+				       "%s: create %s token error 0x%lX",
+				       reader->name, app_name, rv);
+				goto fail;
 			}
+			/* p11card is now bound to some slot */
+			free_p11card = 0;
 		}
 	}
 
 	sc_log(context, "%s: Detection ended", reader->name);
-	return CKR_OK;
+	rv = CKR_OK;
+
+fail:
+	if (free_p11card) {
+		sc_pkcs11_card_free(p11card);
+	}
+
+	return rv;
 }
 
 
 CK_RV
 card_detect_all(void)
 {
-	unsigned int i;
+	unsigned int i, j;
 
 	sc_log(context, "Detect all cards");
 	/* Detect cards in all initialized readers */
 	for (i=0; i< sc_ctx_get_reader_count(context); i++) {
 		sc_reader_t *reader = sc_ctx_get_reader(context, i);
+
 		if (reader->flags & SC_READER_REMOVED) {
-			struct sc_pkcs11_slot *slot;
 			card_removed(reader);
-			while ((slot = reader_get_slot(reader))) {
-				delete_slot(slot);
+			/* do not remove slots related to this reader which would be
+			 * possible according to PKCS#11 2.20 and later, because NSS can't
+			 * handle a shrinking slot list
+			 * https://bugzilla.mozilla.org/show_bug.cgi?id=1613632 */
+
+			/* Instead, remove the relation between reader and slot */
+			for (j = 0; j<list_size(&virtual_slots); j++) {
+				sc_pkcs11_slot_t *slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, j);
+				if (slot->reader == reader) {
+					slot->reader = NULL;
+				}
 			}
-			_sc_delete_reader(context, reader);
-			i--;
 		} else {
-			if (!reader_get_slot(reader))
-				initialize_reader(reader);
-			else
-				card_detect(sc_ctx_get_reader(context, i));
+			/* Locate a slot related to the reader */
+			int found = 0;
+			for (j = 0; j<list_size(&virtual_slots); j++) {
+				sc_pkcs11_slot_t *slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, j);
+				if (slot->reader == reader) {
+					found = 1;
+					break;
+				}
+			}
+			if (!found) {
+				for (j = 0; j < sc_pkcs11_conf.slots_per_card; j++) {
+					CK_RV rv = create_slot(reader);
+					if (rv != CKR_OK)
+						return rv;
+				}
+			}
+			card_detect(reader);
 		}
 	}
 	sc_log(context, "All cards detected");
@@ -392,7 +465,7 @@ CK_RV slot_get_slot(CK_SLOT_ID id, struct sc_pkcs11_slot ** slot)
 
 CK_RV slot_get_token(CK_SLOT_ID id, struct sc_pkcs11_slot ** slot)
 {
-	int rv;
+	CK_RV rv;
 
 	sc_log(context, "Slot(id=0x%lX): get token", id);
 	rv = slot_get_slot(id, slot);
@@ -418,7 +491,8 @@ CK_RV slot_get_token(CK_SLOT_ID id, struct sc_pkcs11_slot ** slot)
 
 CK_RV slot_token_removed(CK_SLOT_ID id)
 {
-	int rv, token_was_present;
+	CK_RV rv;
+	int token_was_present;
 	struct sc_pkcs11_slot *slot;
 	struct sc_pkcs11_object *object;
 
@@ -439,21 +513,24 @@ CK_RV slot_token_removed(CK_SLOT_ID id)
 
 	/* Release framework stuff */
 	if (slot->p11card != NULL) {
-		if (slot->fw_data != NULL &&
-				slot->p11card->framework != NULL && slot->p11card->framework->release_token != NULL) {
+		if (slot->fw_data != NULL && slot->p11card->framework != NULL
+				&& slot->p11card->framework->release_token != NULL) {
 			slot->p11card->framework->release_token(slot->p11card, slot->fw_data);
 			slot->fw_data = NULL;
 		}
+		slot->p11card = NULL;
 	}
 
 	/* Reset relevant slot properties */
 	slot->slot_info.flags &= ~CKF_TOKEN_PRESENT;
 	slot->login_user = -1;
-	slot->p11card = NULL;
+	pop_all_login_states(slot);
 
 	if (token_was_present)
 		slot->events = SC_EVENT_CARD_REMOVED;
 
+	memset(&slot->token_info, 0, sizeof slot->token_info);
+
 	return CKR_OK;
 }
 
@@ -466,7 +543,9 @@ CK_RV slot_find_changed(CK_SLOT_ID_PTR idp, int mask)
 	card_detect_all();
 	for (i=0; i<list_size(&virtual_slots); i++) {
 		sc_pkcs11_slot_t *slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, i);
-		sc_log(context, "slot 0x%lx token: %d events: 0x%02X",slot->id, (slot->slot_info.flags & CKF_TOKEN_PRESENT), slot->events);
+		sc_log(context, "slot 0x%lx token: %lu events: 0x%02X",
+		       slot->id, (slot->slot_info.flags & CKF_TOKEN_PRESENT),
+		       slot->events);
 		if ((slot->events & SC_EVENT_CARD_INSERTED)
 				&& !(slot->slot_info.flags & CKF_TOKEN_PRESENT)) {
 			/* If a token has not been initialized, clear the inserted event */
