@@ -24,6 +24,10 @@
 #include <config.h>
 #endif
 
+#ifndef FIX_UNUSED
+#define FIX_UNUSED(X) (void) (X) /* avoid warnings for unused params */
+#endif
+
 #ifdef ENABLE_OPENSSL   /* empty file without openssl */
 
 #include <stdlib.h>
@@ -93,7 +97,7 @@ iasecc_pkcs15_delete_file(struct sc_pkcs15_card *p15card, struct sc_profile *pro
 	rv = sc_pkcs15init_authenticate(profile, p15card, df, SC_AC_OP_DELETE);
 	card->caps = caps;
 
-	LOG_TEST_RET(ctx, rv, "Cannnot authenticate SC_AC_OP_DELETE");
+	LOG_TEST_RET(ctx, rv, "Cannot authenticate SC_AC_OP_DELETE");
 
 	memset(&path, 0, sizeof(path));
 	path.type = SC_PATH_TYPE_FILE_ID;
@@ -335,7 +339,8 @@ iasecc_file_convert_acls(struct sc_context *ctx, struct sc_profile *profile, str
 		if (acl)   {
 			switch (acl->method)   {
 			case SC_AC_IDA:
-				LOG_TEST_RET(ctx, SC_ERROR_NOT_SUPPORTED, "'IDA' not actually supported");
+				sc_log(ctx, "'IDA' not actually supported");
+				return SC_ERROR_NOT_SUPPORTED;
 			case SC_AC_SCB:
 				if ((acl->key_ref & IASECC_SCB_METHOD_MASK) == IASECC_SCB_METHOD_USER_AUTH)   {
 					acl->method = SC_AC_SEN;
@@ -373,7 +378,7 @@ iasecc_sdo_set_key_acls_from_profile(struct sc_profile *profile, struct sc_card
 
 	/* Get ACLs from profile template */
 	rv = sc_profile_get_file(profile, template, &file);
-	LOG_TEST_RET(ctx, rv, "IasEcc: cannot instanciate private key file");
+	LOG_TEST_RET(ctx, rv, "IasEcc: cannot instantiate private key file");
 
 	/* Convert PKCS15 ACLs to SE ACLs */
 	rv = iasecc_file_convert_acls(ctx, profile, file);
@@ -472,7 +477,7 @@ iasecc_sdo_allocate_prvkey(struct sc_profile *profile, struct sc_card *card,
 		sdo->docp.non_repudiation.value = calloc(1, 1);
 		if (!sdo->docp.non_repudiation.value)
 			LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
-		sdo->docp.non_repudiation.tag = IASECC_DOCP_TAG_NON_REPUDATION;
+		sdo->docp.non_repudiation.tag = IASECC_DOCP_TAG_NON_REPUDIATION;
 		sdo->docp.non_repudiation.size = 1;
 
 		sdo->data.prv_key.compulsory.value = calloc(1, 1);
@@ -495,7 +500,7 @@ iasecc_sdo_allocate_prvkey(struct sc_profile *profile, struct sc_card *card,
 		If present, this attribute has to be the same in the 'GENERATE KEY' template data.
 */
 		if (!(key_info->access_flags & SC_PKCS15_PRKEY_ACCESS_LOCAL) && (key_info->usage & SC_PKCS15_PRKEY_USAGE_NONREPUDIATION))
-			sc_log(ctx, "Non fatal error: NON_REPUDATION can be used only for the localy generated keys");
+			sc_log(ctx, "Non fatal error: NON_REPUDIATION can be used only for the locally generated keys");
 
 		if ((key_info->access_flags & SC_PKCS15_PRKEY_ACCESS_LOCAL)
 				&& (key_info->usage & SC_PKCS15_PRKEY_USAGE_SIGN)
@@ -777,7 +782,7 @@ iasecc_pkcs15_fix_file_access(struct sc_pkcs15_card *p15card, struct sc_file *fi
 }
 
 
-static int
+int
 iasecc_pkcs15_encode_supported_algos(struct sc_pkcs15_card *p15card, struct sc_pkcs15_object *object)
 {
 	struct sc_context *ctx = p15card->card->ctx;
@@ -796,7 +801,8 @@ iasecc_pkcs15_encode_supported_algos(struct sc_pkcs15_card *p15card, struct sc_p
 			LOG_TEST_RET(ctx, rv, "cannot add supported algorithm DECIPHER:CKM_RSA_PKCS");
 		}
 
-		if (prkey_info->usage & SC_PKCS15_PRKEY_USAGE_SIGN)   {
+		if (prkey_info->usage & (SC_PKCS15_PRKEY_USAGE_SIGN |
+		                         SC_PKCS15_PRKEY_USAGE_NONREPUDIATION))   {
 			if (prkey_info->usage & SC_PKCS15_PRKEY_USAGE_NONREPUDIATION)   {
 				algo = sc_pkcs15_get_supported_algo(p15card, SC_PKCS15_ALGO_OP_COMPUTE_SIGNATURE, CKM_SHA1_RSA_PKCS);
 				rv = sc_pkcs15_add_supported_algo_ref(object, algo);
@@ -1136,7 +1142,7 @@ iasecc_pkcs15_generate_key(struct sc_profile *profile, sc_pkcs15_card_t *p15card
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 
-	/* TODO: Check if native IAS middleware accepts the meaningfull path value. */
+	/* TODO: Check if native IAS middleware accepts the meaningful path value. */
 	rv = sc_profile_get_parent(profile, "private-key", &file);
 	LOG_TEST_RET(ctx, rv, "IasEcc: cannot get private key parent file");
 
@@ -1484,8 +1490,11 @@ iasecc_md_gemalto_unset_default(struct sc_pkcs15_card *p15card, struct sc_profil
 	rv = sc_pkcs15_read_data_object(p15card, (struct sc_pkcs15_data_info *)data_obj->data, &dod);
 	LOG_TEST_RET(ctx, rv, "Cannot read from 'CSP/'Default Key Container'");
 
-	if (guid_len != dod->data_len || memcmp(guid, dod->data, guid_len))
+	if (guid_len != dod->data_len || memcmp(guid, dod->data, guid_len)) {
+		sc_pkcs15_free_data_object(dod);
 		LOG_FUNC_RETURN(ctx, SC_SUCCESS);
+	}
+	sc_pkcs15_free_data_object(dod);
 
 	rv = sc_pkcs15_get_objects(p15card, SC_PKCS15_TYPE_PRKEY, key_objs, 32);
 	LOG_TEST_RET(ctx, rv, "Get private key PKCS#15 objects error");
@@ -1659,7 +1668,8 @@ iasecc_store_pubkey(struct sc_pkcs15_card *p15card, struct sc_profile *profile,
 	pubkey_info->usage |= prkey_info->usage & SC_PKCS15_PRKEY_USAGE_DECRYPT ? SC_PKCS15_PRKEY_USAGE_ENCRYPT : 0;
 	pubkey_info->usage |= prkey_info->usage & SC_PKCS15_PRKEY_USAGE_UNWRAP ? SC_PKCS15_PRKEY_USAGE_WRAP : 0;
 
-	iasecc_pkcs15_add_access_rule(object, SC_PKCS15_ACCESS_RULE_MODE_READ, NULL);
+	rv = iasecc_pkcs15_add_access_rule(object, SC_PKCS15_ACCESS_RULE_MODE_READ, NULL);
+	LOG_TEST_RET(ctx, rv, "Too many access rules");
 
 	memcpy(&pubkey_info->algo_refs[0], &prkey_info->algo_refs[0], sizeof(pubkey_info->algo_refs));
 
@@ -1876,4 +1886,19 @@ sc_pkcs15init_get_iasecc_ops(void)
 	return &sc_pkcs15init_iasecc_operations;
 }
 
+#else /* ENABLE_OPENSSL */
+#include "../libopensc/log.h"
+#include "pkcs15-init.h"
+
+int
+iasecc_pkcs15_encode_supported_algos(struct sc_pkcs15_card *p15card, struct sc_pkcs15_object *object)
+{
+	struct sc_context *ctx = p15card->card->ctx;
+	FIX_UNUSED(object);
+
+	LOG_FUNC_CALLED(ctx);
+	sc_log(ctx, "OpenSC was built without OpenSSL support: skipping");
+	LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_IMPLEMENTED);
+}
+
 #endif /* ENABLE_OPENSSL */
