@@ -20,6 +20,7 @@
  */
 
 #include "p11test_case_common.h"
+#include "../../libopensc/sc-ossl-compat.h"
 
 char name_buffer[11];
 char flag_buffer[11];
@@ -114,7 +115,7 @@ add_supported_mechs(test_cert_t *o)
 		if (token.num_rsa_mechs > 0 ) {
 			/* Get supported mechanisms by token */
 			o->num_mechs = token.num_rsa_mechs;
-			for (i = 0; i <= token.num_rsa_mechs; i++) {
+			for (i = 0; i < token.num_rsa_mechs; i++) {
 				o->mechs[i].mech = token.rsa_mechs[i].mech;
 				o->mechs[i].result_flags = 0;
 				o->mechs[i].usage_flags =
@@ -131,7 +132,7 @@ add_supported_mechs(test_cert_t *o)
 	} else if (o->type == EVP_PK_EC) {
 		if (token.num_ec_mechs > 0 ) {
 			o->num_mechs = token.num_ec_mechs;
-			for (i = 0; i <= token.num_ec_mechs; i++) {
+			for (i = 0; i < token.num_ec_mechs; i++) {
 				o->mechs[i].mech = token.ec_mechs[i].mech;
 				o->mechs[i].result_flags = 0;
 				o->mechs[i].usage_flags =
@@ -144,6 +145,38 @@ add_supported_mechs(test_cert_t *o)
 			o->mechs[0].result_flags = 0;
 			o->mechs[0].usage_flags = CKF_SIGN | CKF_VERIFY;
 		}
+	} else if (o->type == EVP_PKEY_ED25519) {
+		if (token.num_ed_mechs > 0 ) {
+			o->num_mechs = token.num_ed_mechs;
+			for (i = 0; i < token.num_ed_mechs; i++) {
+				o->mechs[i].mech = token.ed_mechs[i].mech;
+				o->mechs[i].result_flags = 0;
+				o->mechs[i].usage_flags =
+					token.ed_mechs[i].usage_flags;
+			}
+		} else {
+			/* Use the default list */
+			o->num_mechs = 1;
+			o->mechs[0].mech = CKM_EDDSA;
+			o->mechs[0].result_flags = 0;
+			o->mechs[0].usage_flags = CKF_SIGN | CKF_VERIFY;
+		}
+	} else if (o->type == EVP_PKEY_X25519) {
+		if (token.num_montgomery_mechs > 0 ) {
+			o->num_mechs = token.num_montgomery_mechs;
+			for (i = 0; i < token.num_ed_mechs; i++) {
+				o->mechs[i].mech = token.montgomery_mechs[i].mech;
+				o->mechs[i].result_flags = 0;
+				o->mechs[i].usage_flags =
+					token.montgomery_mechs[i].usage_flags;
+			}
+		} else {
+			/* Use the default list */
+			o->num_mechs = 1;
+			o->mechs[0].mech = CKM_ECDH1_DERIVE;
+			o->mechs[0].result_flags = 0;
+			o->mechs[0].usage_flags = CKF_DERIVE;
+		}
 	}
 }
 
@@ -176,8 +209,8 @@ int callback_certificates(test_certs_t *objects,
 
 	if (EVP_PKEY_base_id(evp) == EVP_PKEY_RSA) {
 		/* Extract public RSA key */
-		RSA *rsa = EVP_PKEY_get0_RSA(evp);
-		if ((o->key.rsa = RSAPublicKey_dup(rsa)) == NULL) {
+		const RSA *rsa = EVP_PKEY_get0_RSA(evp);
+		if ((o->key.rsa = RSAPublicKey_dup((RSA *)rsa)) == NULL) {
 			fail_msg("RSAPublicKey_dup failed");
 			return -1;
 		}
@@ -186,7 +219,7 @@ int callback_certificates(test_certs_t *objects,
 
 	} else if (EVP_PKEY_base_id(evp) == EVP_PKEY_EC) {
 		/* Extract public EC key */
-		EC_KEY *ec = EVP_PKEY_get0_EC_KEY(evp);
+		const EC_KEY *ec = EVP_PKEY_get0_EC_KEY(evp);
 		if ((o->key.ec = EC_KEY_dup(ec)) == NULL) {
 			fail_msg("EC_KEY_dup failed");
 			return -1;
@@ -399,8 +432,124 @@ int callback_public_keys(test_certs_t *objects,
 			EC_KEY_set_group(o->key.ec, ecgroup);
 			o->bits = EC_GROUP_get_degree(ecgroup);
 		}
+	} else if (o->key_type == CKK_EC_EDWARDS
+		|| o->key_type == CKK_EC_MONTGOMERY) {
+		EVP_PKEY *key = NULL;
+		ASN1_PRINTABLESTRING *curve = NULL;
+		ASN1_OBJECT *obj = NULL;
+		const unsigned char *a;
+		ASN1_OCTET_STRING *os;
+		int evp_type;
+
+		a = template[6].pValue;
+		if (d2i_ASN1_PRINTABLESTRING(&curve, &a, (long)template[6].ulValueLen) != NULL) {
+			switch (o->key_type) {
+			case CKK_EC_EDWARDS:
+				if (strcmp((char *)curve->data, "edwards25519")) {
+					debug_print(" [WARN %s ] Unknown curve name. "
+						" expected edwards25519, got %s", o->id_str, curve->data);
+				}
+				evp_type = EVP_PKEY_ED25519;
+				break;
+			case CKK_EC_MONTGOMERY:
+				if (strcmp((char *)curve->data, "curve25519")) {
+					debug_print(" [WARN %s ] Unknown curve name. "
+						" expected curve25519, got %s", o->id_str, curve->data);
+				}
+				evp_type = EVP_PKEY_X25519;
+				break;
+			default:
+				debug_print(" [WARN %s ] Unknown key type %lu", o->id_str, o->key_type);
+				return -1;
+			}
+			ASN1_PRINTABLESTRING_free(curve);
+		} else if (d2i_ASN1_OBJECT(&obj, &a, (long)template[6].ulValueLen) != NULL) {
+			int nid = OBJ_obj2nid(obj);
+			ASN1_OBJECT_free(obj);
+
+			switch (o->key_type) {
+			case CKK_EC_EDWARDS:
+				if (nid != NID_ED25519) {
+					debug_print(" [WARN %s ] Unknown OID. "
+						" expected NID_ED25519 (%d), got %d", o->id_str, NID_ED25519, nid);
+				}
+				evp_type = EVP_PKEY_ED25519;
+				break;
+			case CKK_EC_MONTGOMERY:
+				if (nid != NID_X25519) {
+					debug_print(" [WARN %s ] Unknown OID. "
+						" expected NID_X25519 (%d), got %d", o->id_str, NID_X25519, nid);
+				}
+				evp_type = EVP_PKEY_X25519;
+				break;
+			default:
+				debug_print(" [WARN %s ] Unknown key type %lu", o->id_str, o->key_type);
+				return -1;
+			}
+		} else {
+			debug_print(" [WARN %s ] Failed to convert EC_PARAMS"
+				" to curve name or object id", o->id_str);
+			return -1;
+		}
+
+		/* PKCS#11-compliant modules should return ASN1_OCTET_STRING */
+		a = template[7].pValue;
+		os = d2i_ASN1_OCTET_STRING(NULL, &a, (long)template[7].ulValueLen);
+		if (!os) {
+			debug_print(" [WARN %s ] Can not decode EC_POINT", o->id_str);
+			return -1;
+		}
+		if (os->length != 32) {
+			debug_print(" [WARN %s ] Invalid length of EC_POINT value", o->id_str);
+			return -1;
+		}
+		key = EVP_PKEY_new_raw_public_key(evp_type, NULL,
+			(const uint8_t *)os->data,
+			os->length);
+		if (key == NULL) {
+			debug_print(" [WARN %s ] Out of memory", o->id_str);
+			ASN1_STRING_free(os);
+			return -1;
+		}
+		if (o->key.pkey != NULL) {
+			unsigned char *pub = NULL;
+			size_t publen = 0;
+
+			/* TODO check EVP_PKEY type */
+
+			if (EVP_PKEY_get_raw_public_key(o->key.pkey, NULL, &publen) != 1) {
+				debug_print(" [WARN %s ] Can not get size of the key", o->id_str);
+				ASN1_STRING_free(os);
+				return -1;
+			}
+			pub = malloc(publen);
+			if (pub == NULL) {
+				debug_print(" [WARN %s ] Out of memory", o->id_str);
+				ASN1_STRING_free(os);
+				return -1;
+			}
+
+			if (EVP_PKEY_get_raw_public_key(o->key.pkey, pub, &publen) != 1 ||
+				publen != (size_t)os->length ||
+				memcmp(pub, os->data, publen) != 0) {
+				debug_print(" [WARN %s ] Got different public"
+					"key then from the certificate",
+					o->id_str);
+				free(pub);
+				ASN1_STRING_free(os);
+				return -1;
+			}
+			free(pub);
+			EVP_PKEY_free(key);
+			o->verify_public = 1;
+		} else { /* store the public key for future use */
+			o->type = evp_type;
+			o->key.pkey = key;
+			o->bits = 255;
+		}
+		ASN1_STRING_free(os);
 	} else {
-		debug_print(" [WARN %s ] non-RSA, non-EC key. Key type: %02lX",
+		debug_print(" [WARN %s ] unknown key. Key type: %02lX",
 			o->id_str, o->key_type);
 		return -1;
 	}
@@ -577,10 +726,15 @@ void clean_all_objects(test_certs_t *objects) {
 		free(objects->data[i].label);
 		X509_free(objects->data[i].x509);
 		if (objects->data[i].key_type == CKK_RSA &&
-		    objects->data[i].key.rsa != NULL)
+		    objects->data[i].key.rsa != NULL) {
 			RSA_free(objects->data[i].key.rsa);
-		else if (objects->data[i].key.ec != NULL)
+		} else if (objects->data[i].key_type == CKK_EC &&
+			objects->data[i].key.ec != NULL) {
 			EC_KEY_free(objects->data[i].key.ec);
+		} else if (objects->data[i].key_type == CKK_EC_EDWARDS &&
+			objects->data[i].key.pkey != NULL) {
+			EVP_PKEY_free(objects->data[i].key.pkey);
+		}
 	}
 	free(objects->data);
 }
@@ -606,20 +760,30 @@ const char *get_mechanism_name(int mech_id)
 			return "ECDSA";
 		case CKM_ECDSA_SHA1:
 			return "ECDSA_SHA1";
+		case CKM_ECDSA_SHA224:
+			return "ECDSA_SHA224";
 		case CKM_ECDSA_SHA256:
 			return "ECDSA_SHA256";
 		case CKM_ECDSA_SHA384:
 			return "ECDSA_SHA384";
 		case CKM_ECDSA_SHA512:
 			return "ECDSA_SHA512";
+		case CKM_EDDSA:
+			return "EDDSA";
+		case CKM_XEDDSA:
+			return "XEDDSA";
 		case CKM_ECDH1_DERIVE:
 			return "ECDH1_DERIVE";
 		case CKM_ECDH1_COFACTOR_DERIVE:
 			return "ECDH1_COFACTOR_DERIVE";
 		case CKM_EC_KEY_PAIR_GEN:
 			return "EC_KEY_PAIR_GEN";
+		case CKM_EC_EDWARDS_KEY_PAIR_GEN:
+			return "EC_EDWARDS_KEY_PAIR_GEN";
 		case CKM_RSA_PKCS_KEY_PAIR_GEN:
 			return "RSA_PKCS_KEY_PAIR_GEN";
+		case CKM_GENERIC_SECRET_KEY_GEN:
+			return "GENERIC_SECRET_KEY_GEN";
 		case CKM_MD5_RSA_PKCS:
 			return "MD5_RSA_PKCS";
 		case CKM_RIPEMD160_RSA_PKCS:
@@ -640,6 +804,8 @@ const char *get_mechanism_name(int mech_id)
 			return "MD5_HMAC";
 		case CKM_SHA_1_HMAC:
 			return "SHA_1_HMAC";
+		case CKM_SHA224_HMAC:
+			return "SHA224_HMAC";
 		case CKM_SHA256_HMAC:
 			return "SHA256_HMAC";
 		case CKM_SHA384_HMAC:
@@ -648,6 +814,10 @@ const char *get_mechanism_name(int mech_id)
 			return "SHA512_HMAC";
 		case CKM_RSA_PKCS_OAEP:
 			return "RSA_PKCS_OAEP";
+		case CKM_RIPEMD160:
+			return "RIPEMD160";
+		case CKM_GOSTR3411:
+			return "GOSTR3411";
 		case CKM_MD5:
 			return "MD5";
 		case CKM_SHA_1:
@@ -660,6 +830,14 @@ const char *get_mechanism_name(int mech_id)
 			return "SHA384";
 		case CKM_SHA512:
 			return "SHA512";
+		case CKM_SHA3_256:
+			return "SHA3_256";
+		case CKM_SHA3_224:
+			return "SHA3_224";
+		case CKM_SHA3_384:
+			return "SHA3_384";
+		case CKM_SHA3_512:
+			return "SHA3_512";
 		default:
 			sprintf(name_buffer, "0x%.8X", mech_id);
 			return name_buffer;
